"""Tests that perform actions over orphan files."""
import os
import unittest
from random import choice

from pulp_smash import cli, config, utils
from pulp_smash.exceptions import CalledProcessError
from pulp_smash.pulp3.bindings import monitor_task
from pulp_smash.pulp3.utils import (
    delete_version,
    gen_repo,
    get_content,
    get_versions,
)

from pulpcore.tests.functional.api.using_plugin.constants import FILE_CONTENT_NAME
from pulpcore.client.pulpcore import ArtifactsApi
from pulpcore.client.pulpcore import OrphansApi, OrphansCleanupApi
from pulpcore.client.pulpcore.exceptions import ApiException
from pulpcore.client.pulp_file import (
    ApiClient,
    ContentFilesApi,
    RepositoriesFileApi,
    RepositorySyncURL,
    RemotesFileApi,
)
from pulpcore.tests.functional.utils import configuration, core_client
from pulpcore.tests.functional.api.using_plugin.utils import gen_file_remote
from pulpcore.tests.functional.api.using_plugin.utils import (  # noqa:F401
    set_up_module as setUpModule,
)


class DeleteOrphansTestCase(unittest.TestCase):
    """Test whether orphan files can be cleaned up.

    An orphan artifact is an artifact that is not in any content units.
    An orphan content unit is a content unit that is not in any repository
    version.

    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.api_client = ApiClient(configuration)
        cls.cli_client = cli.Client(cls.cfg)
        cls.orphans_api = OrphansApi(core_client)
        cls.orphans_cleanup_api = OrphansCleanupApi(core_client)
        cls.storage = utils.get_pulp_setting(cls.cli_client, "DEFAULT_FILE_STORAGE")
        cls.media_root = utils.get_pulp_setting(cls.cli_client, "MEDIA_ROOT")
        cls.orphan_protection_time = utils.get_pulp_setting(
            cls.cli_client, "ORPHAN_PROTECTION_TIME"
        )

        orphans_response = cls.orphans_cleanup_api.cleanup({"orphan_protection_time": 0})
        monitor_task(orphans_response.task)

    def test_clean_orphan_content_unit(self):
        """Test whether orphaned content units can be cleaned up."""
        repo_api = RepositoriesFileApi(self.api_client)
        remote_api = RemotesFileApi(self.api_client)

        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        body = gen_file_remote()
        remote = remote_api.create(body)
        self.addCleanup(remote_api.delete, remote.pulp_href)

        # Sync the repository.
        self.assertEqual(repo.latest_version_href, f"{repo.pulp_href}versions/0/")
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = repo_api.read(repo.pulp_href)
        content = choice(get_content(repo.to_dict())[FILE_CONTENT_NAME])

        # Create an orphan content unit.
        repo_api.modify(repo.pulp_href, dict(remove_content_units=[content["pulp_href"]]))

        artifacts_api = ArtifactsApi(core_client)

        if self.storage == "pulpcore.app.models.storage.FileSystem":
            # Verify that the artifact is present on disk.
            relative_path = artifacts_api.read(content["artifact"]).file
            artifact_path = os.path.join(self.media_root, relative_path)
            cmd = ("ls", artifact_path)
            self.cli_client.run(cmd, sudo=True)

        file_contents_api = ContentFilesApi(self.api_client)
        # Delete first repo version. The previous removed content unit will be
        # an orphan.
        delete_version(repo, get_versions(repo.to_dict())[1]["pulp_href"])
        content_units = file_contents_api.list().to_dict()["results"]
        content_units_href = [c["pulp_href"] for c in content_units]
        self.assertIn(content["pulp_href"], content_units_href)

        orphans_response = self.orphans_api.delete()
        monitor_task(orphans_response.task)

        content_units = file_contents_api.list().to_dict()["results"]
        content_units_href = [c["pulp_href"] for c in content_units]

        if self.orphan_protection_time == 0:
            self.assertNotIn(content["pulp_href"], content_units_href)

            if self.storage == "pulpcore.app.models.storage.FileSystem":
                # Verify that the artifact was removed from disk.
                with self.assertRaises(CalledProcessError):
                    self.cli_client.run(cmd)

    def test_clean_orphan_artifact(self):
        """Test whether orphan artifacts units can be clean up."""
        repo_api = RepositoriesFileApi(self.api_client)
        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        artifacts_api = ArtifactsApi(core_client)
        artifact = artifacts_api.create(file=__file__)

        if self.storage == "pulpcore.app.models.storage.FileSystem":
            cmd = ("ls", os.path.join(self.media_root, artifact.file))
            self.cli_client.run(cmd, sudo=True)

        orphans_response = self.orphans_api.delete()
        monitor_task(orphans_response.task)

        if self.orphan_protection_time == 0:
            with self.assertRaises(ApiException):
                artifacts_api.read(artifact.pulp_href)

            if self.storage == "pulpcore.app.models.storage.FileSystem":
                with self.assertRaises(CalledProcessError):
                    self.cli_client.run(cmd)


class OrphansCleanUpTestCase(unittest.TestCase):
    """Test the orphan cleanup endpoint.

    An orphan artifact is an artifact that is not in any content units.
    An orphan content unit is a content unit that is not in any repository
    version.

    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.api_client = ApiClient(configuration)
        cls.cli_client = cli.Client(cls.cfg)
        cls.orphans_cleanup_api = OrphansCleanupApi(core_client)
        cls.storage = utils.get_pulp_setting(cls.cli_client, "DEFAULT_FILE_STORAGE")
        cls.media_root = utils.get_pulp_setting(cls.cli_client, "MEDIA_ROOT")

        orphans_response = cls.orphans_cleanup_api.cleanup({"orphan_protection_time": 0})
        monitor_task(orphans_response.task)

    def test_clean_orphan_content_unit(self):
        """Test whether orphaned content units can be cleaned up."""
        repo_api = RepositoriesFileApi(self.api_client)
        remote_api = RemotesFileApi(self.api_client)

        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        body = gen_file_remote()
        remote = remote_api.create(body)
        self.addCleanup(remote_api.delete, remote.pulp_href)

        # Sync the repository.
        self.assertEqual(repo.latest_version_href, f"{repo.pulp_href}versions/0/")
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = repo_api.read(repo.pulp_href)
        content = choice(get_content(repo.to_dict())[FILE_CONTENT_NAME])

        # Create an orphan content unit.
        repo_api.modify(repo.pulp_href, dict(remove_content_units=[content["pulp_href"]]))

        artifacts_api = ArtifactsApi(core_client)

        if self.storage == "pulpcore.app.models.storage.FileSystem":
            # Verify that the artifact is present on disk.
            relative_path = artifacts_api.read(content["artifact"]).file
            artifact_path = os.path.join(self.media_root, relative_path)
            cmd = ("ls", artifact_path)
            self.cli_client.run(cmd, sudo=True)

        file_contents_api = ContentFilesApi(self.api_client)
        # Delete first repo version. The previous removed content unit will be
        # an orphan.
        delete_version(repo, get_versions(repo.to_dict())[1]["pulp_href"])
        content_units = file_contents_api.list().to_dict()["results"]
        content_units_href = [c["pulp_href"] for c in content_units]
        self.assertIn(content["pulp_href"], content_units_href)

        content_before_cleanup = file_contents_api.list().count
        orphans_response = self.orphans_cleanup_api.cleanup({"orphan_protection_time": 10})
        monitor_task(orphans_response.task)

        # assert content was not removed
        content_after_cleanup = file_contents_api.list().count
        self.assertEqual(content_after_cleanup, content_before_cleanup)

        orphans_response = self.orphans_cleanup_api.cleanup({"orphan_protection_time": 0})
        monitor_task(orphans_response.task)

        content_units = file_contents_api.list().to_dict()["results"]
        content_units_href = [c["pulp_href"] for c in content_units]
        self.assertNotIn(content["pulp_href"], content_units_href)

        if self.storage == "pulpcore.app.models.storage.FileSystem":
            # Verify that the artifact was removed from disk.
            with self.assertRaises(CalledProcessError):
                self.cli_client.run(cmd)

    def test_clean_orphan_artifact(self):
        """Test whether orphan artifacts units can be clean up."""
        repo_api = RepositoriesFileApi(self.api_client)
        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        artifacts_api = ArtifactsApi(core_client)
        artifact = artifacts_api.create(file=__file__)

        if self.storage == "pulpcore.app.models.storage.FileSystem":
            cmd = ("ls", os.path.join(self.media_root, artifact.file))
            self.cli_client.run(cmd, sudo=True)

        orphans_response = self.orphans_cleanup_api.cleanup({"orphan_protection_time": 10})
        monitor_task(orphans_response.task)

        # assert artifact was not removed
        artifacts = artifacts_api.list().count
        self.assertEqual(artifacts, 1)

        orphans_response = self.orphans_cleanup_api.cleanup({"orphan_protection_time": 0})
        monitor_task(orphans_response.task)

        with self.assertRaises(ApiException):
            artifacts_api.read(artifact.pulp_href)

        if self.storage == "pulpcore.app.models.storage.FileSystem":
            with self.assertRaises(CalledProcessError):
                self.cli_client.run(cmd)

    def test_clean_specific_orphans(self):
        """Test whether the `content_hrefs` param removes specific orphans but not others"""
        repo_api = RepositoriesFileApi(self.api_client)
        remote_api = RemotesFileApi(self.api_client)

        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        body = gen_file_remote()
        remote = remote_api.create(body)
        self.addCleanup(remote_api.delete, remote.pulp_href)

        # Sync the repository.
        self.assertEqual(repo.latest_version_href, f"{repo.pulp_href}versions/0/")
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = repo_api.read(repo.pulp_href)

        # Create two orphaned content units.
        content_a = get_content(repo.to_dict())[FILE_CONTENT_NAME][0]["pulp_href"]
        content_b = get_content(repo.to_dict())[FILE_CONTENT_NAME][1]["pulp_href"]
        content_to_remove = dict(remove_content_units=[content_a, content_b])
        repo_api.modify(repo.pulp_href, content_to_remove)

        file_contents_api = ContentFilesApi(self.api_client)
        # Delete first repo version. The previous removed content unit will be an orphan.
        delete_version(repo, get_versions(repo.to_dict())[1]["pulp_href"])
        content_units = file_contents_api.list().to_dict()["results"]
        content_units_href = [c["pulp_href"] for c in content_units]
        self.assertIn(content_a, content_units_href)
        self.assertIn(content_b, content_units_href)

        cleanup_dict = {"content_hrefs": [content_a], "orphan_protection_time": 0}
        orphans_response = self.orphans_cleanup_api.cleanup(cleanup_dict)
        monitor_task(orphans_response.task)

        content_units = file_contents_api.list().to_dict()["results"]
        content_units_href = [c["pulp_href"] for c in content_units]
        self.assertNotIn(content_a, content_units_href)
        self.assertIn(content_b, content_units_href)

    def test_clean_specific_orphans_but_no_orphans_specified(self):
        """Test whether the `content_hrefs` param raises a ValidationError with [] as the value"""
        content_hrefs_dict = {"content_hrefs": []}
        self.assertRaises(ApiException, self.orphans_cleanup_api.cleanup, content_hrefs_dict)

    def test_clean_specific_orphans_but_invalid_orphan_specified(self):
        """Test whether the `content_hrefs` param raises a ValidationError with and invalid href"""
        content_hrefs_dict = {"content_hrefs": ["/not/a/valid/content/href"]}
        self.assertRaises(ApiException, self.orphans_cleanup_api.cleanup, content_hrefs_dict)



# Generated by Django 2.2.5 on 2019-09-17 16:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0006_repository_plugin_managed'),
    ]

    operations = [
        migrations.DeleteModel(
            name='ProgressBar',
        ),
        migrations.DeleteModel(
            name='ProgressSpinner',
        ),
    ]



from gettext import gettext as _

from django_filters.rest_framework import DjangoFilterBackend, filters
from drf_spectacular.utils import extend_schema
from rest_framework import mixins, status
from rest_framework.decorators import action
from rest_framework.filters import OrderingFilter
from rest_framework.response import Response
from rest_framework.serializers import ValidationError

from pulpcore.app.models import Task, TaskGroup, Worker
from pulpcore.app.response import OperationPostponedResponse
from pulpcore.app.serializers import (
    AsyncOperationResponseSerializer,
    MinimalTaskSerializer,
    TaskCancelSerializer,
    TaskGroupSerializer,
    PurgeSerializer,
    TaskSerializer,
    WorkerSerializer,
)
from pulpcore.app.tasks import purge
from pulpcore.app.viewsets import BaseFilterSet, NamedModelViewSet, RolesMixin
from pulpcore.app.viewsets.base import DATETIME_FILTER_OPTIONS, NAME_FILTER_OPTIONS
from pulpcore.app.viewsets.custom_filters import (
    HyperlinkRelatedFilter,
    IsoDateTimeFilter,
    ReservedResourcesFilter,
    CreatedResourcesFilter,
)
from pulpcore.constants import TASK_INCOMPLETE_STATES, TASK_STATES, TASK_CHOICES
from pulpcore.tasking.tasks import dispatch
from pulpcore.tasking.util import cancel as cancel_task


class TaskFilter(BaseFilterSet):
    state = filters.ChoiceFilter(choices=TASK_CHOICES)
    worker = HyperlinkRelatedFilter()
    name = filters.CharFilter()
    logging_cid = filters.CharFilter()
    started_at = IsoDateTimeFilter(field_name="started_at")
    finished_at = IsoDateTimeFilter(field_name="finished_at")
    parent_task = HyperlinkRelatedFilter()
    child_tasks = HyperlinkRelatedFilter()
    task_group = HyperlinkRelatedFilter()
    reserved_resources_record = ReservedResourcesFilter()
    created_resources = CreatedResourcesFilter()

    class Meta:
        model = Task
        fields = {
            "state": ["exact", "in"],
            "worker": ["exact", "in"],
            "name": ["contains"],
            "logging_cid": ["exact", "contains"],
            "started_at": DATETIME_FILTER_OPTIONS,
            "finished_at": DATETIME_FILTER_OPTIONS,
            "parent_task": ["exact"],
            "child_tasks": ["exact"],
            "task_group": ["exact"],
            "reserved_resources_record": ["exact"],
            "created_resources": ["exact"],
        }


class TaskViewSet(
    NamedModelViewSet,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    mixins.DestroyModelMixin,
    RolesMixin,
):
    queryset = Task.objects.all()
    endpoint_name = "tasks"
    filterset_class = TaskFilter
    serializer_class = TaskSerializer
    minimal_serializer_class = MinimalTaskSerializer
    filter_backends = (OrderingFilter, DjangoFilterBackend)
    ordering = "-pulp_created"
    queryset_filtering_required_permission = "core.view_task"

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {"action": ["list"], "principal": "authenticated", "effect": "allow"},
            {
                "action": ["retrieve", "my_permissions"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:core.view_task",
            },
            {
                "action": ["destroy"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:core.delete_task",
            },
            {
                "action": ["update", "partial_update"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:core.change_task",
            },
            # 'purge' is filtered by current-user and core.delete_task permissions at the queryset
            # level, and needs no extra protections here
            {
                "action": ["purge"],
                "principal": "authenticated",
                "effect": "allow",
            },
            {
                "action": ["list_roles", "add_role", "remove_role"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:core.manage_roles_task",
            },
        ],
        "creation_hooks": [
            {
                "function": "add_roles_for_object_creator",
                "parameters": {"roles": "core.task_owner"},
            }
        ],
    }
    LOCKED_ROLES = {
        "core.task_owner": {
            "description": "Allow all actions on a task.",
            "permissions": [
                "core.view_task",
                "core.change_task",
                "core.delete_task",
                "core.manage_roles_task",
            ],
        },
        "core.task_viewer": ["core.view_task"],
    }

    @extend_schema(
        description="This operation cancels a task.",
        summary="Cancel a task",
        operation_id="tasks_cancel",
        responses={200: TaskSerializer, 409: TaskSerializer},
    )
    def partial_update(self, request, pk=None, partial=True):
        task = self.get_object()
        if "state" not in request.data:
            raise ValidationError(_("'state' must be provided with the request."))
        if request.data["state"] != "canceled":
            raise ValidationError(_("The only acceptable value for 'state' is 'canceled'."))
        task = cancel_task(task.pk)
        # Check whether task is actually canceled
        http_status = (
            None
            if task.state in [TASK_STATES.CANCELING, TASK_STATES.CANCELED]
            else status.HTTP_409_CONFLICT
        )
        serializer = self.serializer_class(task, context={"request": request})
        return Response(serializer.data, status=http_status)

    def destroy(self, request, pk=None):
        task = self.get_object()
        if task.state in TASK_INCOMPLETE_STATES:
            return Response(status=status.HTTP_409_CONFLICT)
        return super().destroy(request, pk)

    def get_serializer_class(self):
        if self.action == "partial_update":
            return TaskCancelSerializer
        return super().get_serializer_class()

    @extend_schema(
        description=(
            "Trigger an asynchronous task that deletes completed tasks that finished prior"
            " to a specified timestamp (tech-preview, may change in the future)."
        ),
        summary="Purge Completed Tasks",
        operation_id="tasks_purge",
        request=PurgeSerializer,
        responses={202: AsyncOperationResponseSerializer},
    )
    @action(detail=False, methods=["post"])
    def purge(self, request):
        """
        Purge task-records for tasks in 'final' states.
        """
        serializer = PurgeSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        task = dispatch(
            purge, args=[serializer.data["finished_before"], list(serializer.data["states"])]
        )
        return OperationPostponedResponse(task, request)


class TaskGroupFilter(BaseFilterSet):
    class Meta:
        model = TaskGroup
        fields = ()


class TaskGroupViewSet(NamedModelViewSet, mixins.RetrieveModelMixin, mixins.ListModelMixin):
    queryset = TaskGroup.objects.all()
    endpoint_name = "task-groups"
    filterset_class = TaskGroupFilter
    serializer_class = TaskGroupSerializer
    filter_backends = (OrderingFilter, DjangoFilterBackend)
    ordering = "-pulp_created"


class WorkerFilter(BaseFilterSet):
    name = filters.CharFilter()
    last_heartbeat = IsoDateTimeFilter()
    online = filters.BooleanFilter(method="filter_online")
    missing = filters.BooleanFilter(method="filter_missing")

    class Meta:
        model = Worker
        fields = {
            "name": NAME_FILTER_OPTIONS,
            "last_heartbeat": DATETIME_FILTER_OPTIONS,
        }

    def filter_online(self, queryset, name, value):
        online_workers = Worker.objects.online_workers()

        if value:
            return queryset.filter(pk__in=online_workers)
        else:
            return queryset.exclude(pk__in=online_workers)

    def filter_missing(self, queryset, name, value):
        missing_workers = Worker.objects.missing_workers()

        if value:
            return queryset.filter(pk__in=missing_workers)
        else:
            return queryset.exclude(pk__in=missing_workers)


class WorkerViewSet(NamedModelViewSet, mixins.RetrieveModelMixin, mixins.ListModelMixin):
    queryset = Worker.objects.all()
    serializer_class = WorkerSerializer
    endpoint_name = "workers"
    http_method_names = ["get", "options"]
    lookup_value_regex = "[^/]+"
    filterset_class = WorkerFilter



# Generated by Django 2.2.13 on 2020-07-14 21:00

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0040_set_admin_is_staff'),
    ]

    operations = [
        migrations.CreateModel(
            name='AccessPolicy',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('statements', django.contrib.postgres.fields.jsonb.JSONField()),
                ('viewset_name', models.CharField(max_length=128, unique=True)),
                ('permissions_assignment', django.contrib.postgres.fields.jsonb.JSONField()),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# Generated by Django 2.2.3 on 2019-07-24 17:28

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0002_increase_artifact_size_field'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='upload',
            name='completed',
        ),
    ]



import asyncio
import logging

from gettext import gettext as _

from django.conf import settings

from .profiler import ProfilingQueue


log = logging.getLogger(__name__)


class Stage:
    """
    The base class for all Stages API stages.

    To make a stage, inherit from this class and implement :meth:`run` on the subclass.
    """

    def __init__(self):
        self._in_q = None
        self._out_q = None

    def _connect(self, in_q, out_q):
        """
        Connect to queues within a pipeline.

        Args:
            in_q (asyncio.Queue): The stage input queue.
            out_q (asyncio.Queue): The stage output queue.
        """
        self._in_q = in_q
        self._out_q = out_q

    async def __call__(self):
        """
        This coroutine makes the stage callable.

        It calls :meth:`run` and signals the next stage that its work is finished.
        """
        log.debug(_("%(name)s - begin."), {"name": self})
        await self.run()
        await self._out_q.put(None)
        log.debug(_("%(name)s - put end-marker."), {"name": self})

    async def run(self):
        """
        The coroutine that is run as part of this stage.

        Returns:
            The coroutine that runs this stage.

        """
        raise NotImplementedError(_("A plugin writer must implement this method"))

    async def items(self):
        """
        Asynchronous iterator yielding items of :class:`DeclarativeContent` from `self._in_q`.

        The iterator will get instances of :class:`DeclarativeContent` one by one as they get
        available.

        Yields:
            An instance of :class:`DeclarativeContent`

        Examples:
            Used in stages to get d_content instances one by one from `self._in_q`::

                class MyStage(Stage):
                    async def run(self):
                        async for d_content in self.items():
                            # process declarative content
                            await self.put(d_content)

        """
        while True:
            content = await self._in_q.get()
            if content is None:
                break
            log.debug("%(name)s - next: %(content)s.", {"name": self, "content": content})
            yield content

    async def batches(self, minsize=500):
        """
        Asynchronous iterator yielding batches of :class:`DeclarativeContent` from `self._in_q`.

        The iterator will try to get as many instances of
        :class:`DeclarativeContent` as possible without blocking, but
        at least `minsize` instances.

        Args:
            minsize (int): The minimum batch size to yield (unless it is the final batch)

        Yields:
            A list of :class:`DeclarativeContent` instances

        Examples:
            Used in stages to get large chunks of d_content instances from `self._in_q`::

                class MyStage(Stage):
                    async def run(self):
                        async for batch in self.batches():
                            for d_content in batch:
                                # process declarative content
                                await self.put(d_content)

        """
        batch = []
        shutdown = False
        no_block = False
        thaw_queue_event = asyncio.Event()

        def add_to_batch(content):
            nonlocal batch
            nonlocal shutdown
            nonlocal no_block
            nonlocal thaw_queue_event

            if content is None:
                shutdown = True
                log.debug(_("%(name)s - shutdown."), {"name": self})
            else:
                if not content.does_batch:
                    no_block = True
                content._thaw_queue_event = thaw_queue_event
                batch.append(content)

        get_listener = asyncio.ensure_future(self._in_q.get())
        thaw_event_listener = asyncio.ensure_future(thaw_queue_event.wait())
        while not shutdown:
            done, pending = await asyncio.wait(
                [thaw_event_listener, get_listener], return_when=asyncio.FIRST_COMPLETED
            )
            if thaw_event_listener in done:
                thaw_event_listener = asyncio.ensure_future(thaw_queue_event.wait())
                no_block = True
            if get_listener in done:
                content = await get_listener
                add_to_batch(content)
                get_listener = asyncio.ensure_future(self._in_q.get())
            while not shutdown:
                try:
                    content = self._in_q.get_nowait()
                except asyncio.QueueEmpty:
                    break
                else:
                    add_to_batch(content)

            if batch and (len(batch) >= minsize or shutdown or no_block):
                log.debug(
                    _("%(name)s - next batch[%(length)d]."), {"name": self, "length": len(batch)}
                )
                for content in batch:
                    content._thaw_queue_event = None
                thaw_queue_event.clear()
                yield batch
                batch = []
                no_block = False
        thaw_event_listener.cancel()
        get_listener.cancel()

    async def put(self, item):
        """
        Coroutine to pass items to the next stage.

        Args:
            item: A handled instance of :class:`pulpcore.plugin.stages.DeclarativeContent`

        Raises:
            ValueError: When `item` is None.
        """
        if item is None:
            raise ValueError(_("(None) not permitted."))
        await self._out_q.put(item)
        log.debug("{name} - put: {content}".format(name=self, content=item))

    def __str__(self):
        return "[{id}] {name}".format(id=id(self), name=self.__class__.__name__)


async def create_pipeline(stages, maxsize=1):
    """
    A coroutine that builds a Stages API linear pipeline from the list `stages` and runs it.

    Each stage is an instance of a class derived from :class:`pulpcore.plugin.stages.Stage` that
    implements the :meth:`run` coroutine. This coroutine reads asynchronously either from the
    `items()` iterator or the `batches()` iterator and outputs the items with `put()`. Here is an
    example of the simplest stage that only passes data::

        class MyStage(Stage):
            async def run(self):
                async for d_content in self.items():  # Fetch items from the previous stage
                    await self.put(d_content)  # Hand them over to the next stage

    Args:
        stages (list of coroutines): A list of Stages API compatible coroutines.
        maxsize (int): The maximum amount of items a queue between two stages should hold. Optional
            and defaults to 1.

    Returns:
        A single coroutine that can be used to run, wait, or cancel the entire pipeline with.
    Raises:
        ValueError: When a stage instance is specified more than once.
    """
    futures = []
    history = set()
    in_q = None
    for i, stage in enumerate(stages):
        if stage in history:
            raise ValueError(_("Each stage instance must be unique."))
        history.add(stage)
        if i < len(stages) - 1:
            if settings.PROFILE_STAGES_API:
                out_q = ProfilingQueue.make_and_record_queue(stages[i + 1], i + 1, maxsize)
            else:
                out_q = asyncio.Queue(maxsize=maxsize)
        else:
            out_q = None
        stage._connect(in_q, out_q)
        futures.append(asyncio.ensure_future(stage()))
        in_q = out_q

    try:
        await asyncio.gather(*futures)
    except Exception:
        # One of the stages raised an exception, cancel all stages...
        pending = []
        for task in futures:
            if not task.done():
                task.cancel()
                pending.append(task)
        # ...and run until all Exceptions show up
        if pending:
            await asyncio.wait(pending, timeout=60)
        raise


class EndStage(Stage):
    """
    A Stages API stage that drains incoming items and does nothing with the items. This is
    required at the end of all pipelines.

    Without this stage, the `maxsize` of the last stage's `_out_q` could fill up and block the
    entire pipeline.
    """

    async def __call__(self):
        """
        This method drains items from the last queue and drops them.

        Importantly it does not try to put items into the nonexistent next queue.
        """
        # We overwrite __call__ here to avoid trying to put None in `self._out_q`.
        async for _ in self.items():  # noqa
            pass



import logging

import aiohttp
import asyncio
import backoff

from .base import BaseDownloader, DownloadResult
from pulpcore.exceptions import (
    DigestValidationError,
    SizeValidationError,
    TimeoutException,
)


log = logging.getLogger(__name__)


logging.getLogger("backoff").addHandler(logging.StreamHandler())


def http_giveup_handler(exc):
    """
    Inspect a raised exception and determine if we should give up.

    Do not give up when the error is one of the following:

        HTTP 429 - Too Many Requests
        HTTP 5xx - Server errors
        Socket timeout
        TCP disconnect
        Client SSL Error

    Based on the AWS and Google Cloud guidelines:
        https://docs.aws.amazon.com/general/latest/gr/api-retries.html
        https://cloud.google.com/storage/docs/retry-strategy

    Args:
        exc (Exception): The exception to inspect

    Returns:
        True if the download should give up, False otherwise
    """
    if isinstance(exc, aiohttp.ClientResponseError):
        server_error = 500 <= exc.code < 600
        too_many_requests = exc.code == 429
        return not server_error and not too_many_requests

    # any other type of error (pre-filtered by the backoff decorator) shouldn't be fatal
    return False


class HttpDownloader(BaseDownloader):
    """
    An HTTP/HTTPS Downloader built on `aiohttp`.

    This downloader downloads data from one `url` and is not reused.

    The downloader optionally takes a session argument, which is an `aiohttp.ClientSession`. This
    allows many downloaders to share one `aiohttp.ClientSession` which provides a connection pool,
    connection reuse, and keep-alives across multiple downloaders. When creating many downloaders,
    have one session shared by all of your `HttpDownloader` objects.

    A session is optional; if omitted, one session will be created, used for this downloader, and
    then closed when the download is complete. A session that is passed in will not be closed when
    the download is complete.

    If a session is not provided, the one created by HttpDownloader uses non-default timing values.
    Specifically, the "total" timeout is set to None and the "sock_connect" and "sock_read" are both
    5 minutes. For more info on these settings, see the aiohttp docs:
    http://aiohttp.readthedocs.io/en/stable/client_quickstart.html#timeouts Behaviorally, it should
    allow for an active download to be arbitrarily long, while still detecting dead or closed
    sessions even when TCPKeepAlive is disabled.

    If a session is not provided, the one created will force TCP connection closure after each
    request. This is done for compatibility reasons due to various issues related to session
    continuation implementation in various servers.

    `aiohttp.ClientSession` objects allows you to configure options that will apply to all
    downloaders using that session such as auth, timeouts, headers, etc. For more info on these
    options see the `aiohttp.ClientSession` docs for more information:
    http://aiohttp.readthedocs.io/en/stable/client_reference.html#aiohttp.ClientSession

    The `aiohttp.ClientSession` can additionally be configured for SSL configuration by passing in a
    `aiohttp.TCPConnector`. For information on configuring either server or client certificate based
    identity verification, see the aiohttp documentation:
    http://aiohttp.readthedocs.io/en/stable/client.html#ssl-control-for-tcp-sockets

    For more information on `aiohttp.BasicAuth` objects, see their docs:
    http://aiohttp.readthedocs.io/en/stable/client_reference.html#aiohttp.BasicAuth

    Synchronous Download::

        downloader = HttpDownloader('http://example.com/')
        result = downloader.fetch()

    Parallel Download::

        download_coroutines = [
            HttpDownloader('http://example.com/').run(),
            HttpDownloader('http://pulpproject.org/').run(),
        ]

        loop = asyncio.get_event_loop()
        done, not_done = loop.run_until_complete(asyncio.wait(download_coroutines))

        for task in done:
            try:
                task.result()  # This is a DownloadResult
            except Exception as error:
                pass  # fatal exceptions are raised by result()

    The HTTPDownloaders contain automatic retry logic if the server responds with HTTP 429 response.
    The coroutine will automatically retry 10 times with exponential backoff before allowing a
    final exception to be raised.

    Attributes:
        session (aiohttp.ClientSession): The session to be used by the downloader.
        auth (aiohttp.BasicAuth): An object that represents HTTP Basic Authorization or None
        proxy (str): An optional proxy URL or None
        proxy_auth (aiohttp.BasicAuth): An optional object that represents proxy HTTP Basic
            Authorization or None
        headers_ready_callback (callable): An optional callback that accepts a single dictionary
            as its argument. The callback will be called when the response headers are
            available. The dictionary passed has the header names as the keys and header values
            as its values. e.g. `{'Transfer-Encoding': 'chunked'}`. This can also be None.

    This downloader also has all of the attributes of
    :class:`~pulpcore.plugin.download.BaseDownloader`
    """

    def __init__(
        self,
        url,
        session=None,
        auth=None,
        proxy=None,
        proxy_auth=None,
        headers_ready_callback=None,
        headers=None,
        throttler=None,
        max_retries=0,
        **kwargs,
    ):
        """
        Args:
            url (str): The url to download.
            session (aiohttp.ClientSession): The session to be used by the downloader. (optional) If
                not specified it will open the session and close it
            auth (aiohttp.BasicAuth): An object that represents HTTP Basic Authorization (optional)
            proxy (str): An optional proxy URL.
            proxy_auth (aiohttp.BasicAuth): An optional object that represents proxy HTTP Basic
                Authorization.
            headers_ready_callback (callable): An optional callback that accepts a single dictionary
                as its argument. The callback will be called when the response headers are
                available. The dictionary passed has the header names as the keys and header values
                as its values. e.g. `{'Transfer-Encoding': 'chunked'}`
            headers (dict): Headers to be submitted with the request.
            throttler (asyncio_throttle.Throttler): Throttler for asyncio.
            max_retries (int): The maximum number of times to retry a download upon failure.
            kwargs (dict): This accepts the parameters of
                :class:`~pulpcore.plugin.download.BaseDownloader`.
        """
        if session:
            self.session = session
            self._close_session_on_finalize = False
        else:
            timeout = aiohttp.ClientTimeout(total=None, sock_connect=600, sock_read=600)
            conn = aiohttp.TCPConnector({"force_close": True})
            self.session = aiohttp.ClientSession(connector=conn, timeout=timeout, headers=headers)
            self._close_session_on_finalize = True
        self.auth = auth
        self.proxy = proxy
        self.proxy_auth = proxy_auth
        self.headers_ready_callback = headers_ready_callback
        self.download_throttler = throttler
        self.max_retries = max_retries
        super().__init__(url, **kwargs)

    def raise_for_status(self, response):
        """
        Raise error if aiohttp response status is >= 400 and not silenced.

        Args:
            response (aiohttp.ClientResponse): The response to handle.

        Raises:
               aiohttp.ClientResponseError: When the response status is >= 400.
        """
        response.raise_for_status()

    async def _handle_response(self, response):
        """
        Handle the aiohttp response by writing it to disk and calculating digests

        Args:
            response (aiohttp.ClientResponse): The response to handle.

        Returns:
             DownloadResult: Contains information about the result. See the DownloadResult docs for
                 more information.
        """
        if self.headers_ready_callback:
            await self.headers_ready_callback(response.headers)
        while True:
            chunk = await response.content.read(1048576)  # 1 megabyte
            if not chunk:
                await self.finalize()
                break  # the download is done
            await self.handle_data(chunk)
        return DownloadResult(
            path=self.path,
            artifact_attributes=self.artifact_attributes,
            url=self.url,
            headers=response.headers,
        )

    async def run(self, extra_data=None):
        """
        Run the downloader with concurrency restriction and retry logic.

        This method acquires `self.semaphore` before calling the actual download implementation
        contained in `_run()`. This ensures that the semaphore stays acquired even as the `backoff`
        wrapper around `_run()`, handles backoff-and-retry logic.

        Args:
            extra_data (dict): Extra data passed to the downloader.

        Returns:
            :class:`~pulpcore.plugin.download.DownloadResult` from `_run()`.

        """
        retryable_errors = (
            aiohttp.ClientConnectorSSLError,
            aiohttp.ClientConnectorError,
            aiohttp.ClientOSError,
            aiohttp.ClientPayloadError,
            aiohttp.ClientResponseError,
            aiohttp.ServerDisconnectedError,
            TimeoutError,
            TimeoutException,
            DigestValidationError,
            SizeValidationError,
        )

        async with self.semaphore:

            @backoff.on_exception(
                backoff.expo,
                retryable_errors,
                max_tries=self.max_retries + 1,
                giveup=http_giveup_handler,
            )
            async def download_wrapper():
                self._ensure_no_broken_file()
                try:
                    return await self._run(extra_data=extra_data)
                except asyncio.TimeoutError:
                    raise TimeoutException(self.url)
                except aiohttp.ClientHttpProxyError as e:
                    log.error(
                        "Proxy {!r} rejected connection request during a request to "
                        "{!r}, status={}, message={!r}".format(
                            e.request_info.real_url,
                            e.request_info.url,
                            e.status,
                            e.message,
                        )
                    )
                    raise e

            return await download_wrapper()

    async def _run(self, extra_data=None):
        """
        Download, validate, and compute digests on the `url`. This is a coroutine.

        This method is externally wrapped with backoff-and-retry behavior for some errors.
        It retries with exponential backoff some number of times before allowing a final
        exception to be raised.

        This method provides the same return object type and documented in
        :meth:`~pulpcore.plugin.download.BaseDownloader._run`.

        Args:
            extra_data (dict): Extra data passed by the downloader.
        """
        if self.download_throttler:
            await self.download_throttler.acquire()
        async with self.session.get(
            self.url, proxy=self.proxy, proxy_auth=self.proxy_auth, auth=self.auth
        ) as response:
            self.raise_for_status(response)
            to_return = await self._handle_response(response)
            await response.release()
        if self._close_session_on_finalize:
            await self.session.close()
        return to_return

    def _ensure_no_broken_file(self):
        """Upon retry reset writer back to None to get a fresh file."""
        if self._writer is not None:
            self._writer.delete = True
            self._writer.close()
            self._writer = None



# Generated by Django 2.2.7 on 2019-11-11 16:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0013_repository_pulp_type'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='repository',
            name='plugin_managed',
        ),
    ]



from gettext import gettext as _

from pulpcore.exceptions import PulpException


class ValidationError(PulpException):
    """
    A base class for all Validation Errors.
    """

    pass


class DigestValidationError(ValidationError):
    """
    Raised when a file fails to validate a digest checksum.
    """

    def __init__(self, *args):
        super().__init__("PLP0003")
        if args:
            self.url = args[0]

    def __str__(self):
        if hasattr(self, "url"):
            return _("A file located at the url {} failed validation due to checksum.").format(
                self.url
            )
        return _("A file failed validation due to checksum.")


class SizeValidationError(ValidationError):
    """
    Raised when a file fails to validate a size checksum.
    """

    def __init__(self, *args):
        super().__init__("PLP0004")
        if args:
            self.url = args[0]

    def __str__(self):
        if hasattr(self, "url"):
            return _("A file located at the url {} failed validation due to size.").format(self.url)
        return _("A file failed validation due to size.")


class MissingDigestValidationError(Exception):
    """
    Raised when attempting to save() an Artifact with an incomplete set of checksums.
    """

    pass


class UnsupportedDigestValidationError(Exception):
    """
    Raised when an attempt is made to use a checksum-type that is not enabled/available.
    """

    pass



from unittest import TestCase
from pulpcore.app.serializers import PulpExportSerializer


class TestPulpExportSerializer(TestCase):
    def test_validate_no_params(self):
        data = {}
        serializer = PulpExportSerializer(data=data)
        self.assertTrue(serializer.is_valid())

    def test_validate_bad_param_values(self):
        data = {"full": "bar", "dry_run": 0}
        serializer = PulpExportSerializer(data=data)
        self.assertFalse(serializer.is_valid())

    def test_bad_params(self):
        data = {"baz": "bar"}
        serializer = PulpExportSerializer(data=data)
        self.assertFalse(serializer.is_valid())

    def test_read_only_params(self):
        data = {"full": True, "dry_run": False, "output_file_info": {"bar": "blech"}}
        serializer = PulpExportSerializer(data=data)
        self.assertTrue(serializer.is_valid())

        with self.assertRaises(AttributeError):
            serializer.output_file_info["bar"]

    def test_chunk_size(self):
        # positive tests
        # bytes
        data = {"chunk_size": "100B"}
        serializer = PulpExportSerializer(data=data)
        self.assertTrue(serializer.is_valid())
        self.assertEqual(100, serializer.validated_data["chunk_size"])

        # kilobytes
        data = {"chunk_size": "100KB"}
        serializer = PulpExportSerializer(data=data)
        self.assertTrue(serializer.is_valid())
        self.assertEqual(100 * 1024, serializer.validated_data["chunk_size"])

        # megabytes
        data = {"chunk_size": "100MB"}
        serializer = PulpExportSerializer(data=data)
        self.assertTrue(serializer.is_valid())
        self.assertEqual(100 * 1024 * 1024, serializer.validated_data["chunk_size"])

        # gigabytes
        data = {"chunk_size": "100GB"}
        serializer = PulpExportSerializer(data=data)
        self.assertTrue(serializer.is_valid())
        self.assertEqual(100 * 1024 * 1024 * 1024, serializer.validated_data["chunk_size"])

        # terabytes
        data = {"chunk_size": "1TB"}
        serializer = PulpExportSerializer(data=data)
        self.assertTrue(serializer.is_valid())
        self.assertEqual(1 * 1024 * 1024 * 1024 * 1024, serializer.validated_data["chunk_size"])

        # float-units
        data = {"chunk_size": "2.4GB"}
        serializer = PulpExportSerializer(data=data)
        self.assertTrue(serializer.is_valid())
        self.assertEqual(int(2.4 * 1024 * 1024 * 1024), serializer.validated_data["chunk_size"])

        # negative tests
        # no units
        data = {"chunk_size": "100"}
        serializer = PulpExportSerializer(data=data)
        self.assertFalse(serializer.is_valid())

        # not-a-number
        data = {"chunk_size": "bazMB"}
        serializer = PulpExportSerializer(data=data)
        serializer.is_valid()

        # non-positive
        data = {"chunk_size": "0GB"}
        serializer = PulpExportSerializer(data=data)
        self.assertFalse(serializer.is_valid())

        # non-positive
        data = {"chunk_size": "-10KB"}
        serializer = PulpExportSerializer(data=data)
        self.assertFalse(serializer.is_valid())

        # too many terabytes
        data = {"chunk_size": "100TB"}
        serializer = PulpExportSerializer(data=data)
        self.assertFalse(serializer.is_valid())

        # morbidly many megabytes
        data = {"chunk_size": "10000000000000M"}
        serializer = PulpExportSerializer(data=data)
        self.assertFalse(serializer.is_valid())



import enum
import json

from functools import wraps

from aiohttp.web import FileResponse, Response, HTTPSuccessful, Request
from aiohttp.web_exceptions import HTTPFound

from redis import ConnectionError
from aioredis import ConnectionError as AConnectionError

from pulpcore.app.settings import settings
from pulpcore.app.redis_connection import (
    get_redis_connection,
    get_async_redis_connection,
)

DEFAULT_EXPIRES_TTL = settings.CACHE_SETTINGS["EXPIRES_TTL"]


class CacheKeys(enum.Enum):
    """Available keys to construct the index key for cache entry."""

    path = "path"
    host = "host"
    method = "method"


def connection_error_wrapper(func):
    """A decorator that enables sync functions which use Redis to swallow connection errors."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        """Handle connection errors, specific to the sync context, raised by the Redis client."""
        try:
            return func(*args, **kwargs)
        except ConnectionError:
            return None

    return wrapper


def aconnection_error_wrapper(func):
    """A decorator that enables async functions which use Redis to swallow connection errors."""

    @wraps(func)
    async def wrapper(*args, **kwargs):
        """Handle connection errors, specific to the async context, raised by the Redis client."""
        try:
            return await func(*args, **kwargs)
        except AConnectionError:
            return None

    return wrapper


class Cache:
    """Base class for Pulp's cache"""

    default_base_key = "PULP_CACHE"
    default_expires_ttl = DEFAULT_EXPIRES_TTL

    def __init__(self):
        """Creates synchronous cache instance"""
        self.redis = get_redis_connection()

    @connection_error_wrapper
    def get(self, key, base_key=None):
        """Gets cached entry of key"""
        base_key = base_key or self.default_base_key
        if key is None:
            return self.redis.hgetall(base_key)
        return self.redis.hget(base_key, key)

    @connection_error_wrapper
    def set(self, key, value, expires=None, base_key=None):
        """Sets the cached entry at key"""
        base_key = base_key or self.default_base_key
        ret = self.redis.hset(base_key, key, value)
        if expires:
            self.redis.expire(base_key, expires)
        return ret

    @connection_error_wrapper
    def exists(self, key=None, base_key=None):
        """Checks if cached entries exist"""
        if not base_key and base_key is not None:
            return False  # Failsafe for passing empty list/str
        base_key = base_key or self.default_base_key
        if key:
            return self.redis.hexists(base_key, key)
        else:
            if isinstance(base_key, str):
                base_key = [base_key]
            return self.redis.exists(*base_key)

    @connection_error_wrapper
    def delete(self, key=None, base_key=None):
        """
        Deletes the cached entry at base_key: key

        If only base_key is supplied then delete all entries under that base_key
        key can be a list to delete multiple entries under a base_key
        base_key can be a list to delete multiple sets of entries
        key and base_key should not both be lists
        """
        base_key = base_key or self.default_base_key
        if key:
            return self.redis.hdel(base_key, key)
        if isinstance(base_key, str):
            base_key = [base_key]
        return self.redis.delete(*base_key)


class AsyncCache:
    """Base class for asynchronous Pulp Cache"""

    default_base_key = "PULP_CACHE"
    default_expires_ttl = DEFAULT_EXPIRES_TTL

    def __init__(self):
        """Creates asynchronous cache instance"""
        self.redis = get_async_redis_connection()

    @aconnection_error_wrapper
    async def get(self, key, base_key=None):
        """Gets cached entry of key"""
        base_key = base_key or self.default_base_key
        if key is None:
            return await self.redis.hgetall(base_key)
        return await self.redis.hget(base_key, key)

    @aconnection_error_wrapper
    async def set(self, key, value, expires=None, base_key=None):
        """Sets the cached entry at key"""
        base_key = base_key or self.default_base_key
        ret = await self.redis.hset(base_key, key, value)
        if expires:
            await self.redis.expire(base_key, expires)
        return ret

    @aconnection_error_wrapper
    async def exists(self, key=None, base_key=None):
        """Checks if cached entries exist"""
        if not base_key and base_key is not None:
            return False  # Failsafe for passing empty list/str
        base_key = base_key or self.default_base_key
        if key:
            return await self.redis.hexists(base_key, key)
        else:
            if isinstance(base_key, str):
                base_key = [base_key]
            return await self.redis.exists(*base_key)

    @aconnection_error_wrapper
    async def delete(self, key=None, base_key=None):
        """
        Deletes the cached entry at base_key: key

        If only base_key is supplied then delete all entries under that base_key
        key can be a list to delete multiple entries under a base_key
        base_key can be a list to delete multiple sets of entries
        key and base_key should not both be lists
        """
        base_key = base_key or self.default_base_key
        if key:
            return await self.redis.hdel(base_key, key)
        if isinstance(base_key, str):
            base_key = [base_key]
        return await self.redis.delete(*base_key)


class ContentCache(AsyncCache):
    """Cache object meant to be used for the content app"""

    RESPONSE_TYPES = {
        "FileResponse": FileResponse,
        "Response": Response,
        "Redirect": HTTPFound,
    }

    def __init__(self, base_key=None, expires_ttl=None, keys=None, auth=None):
        """
        Initiates a cache instance to be used for dealing with an aiohttp server

        Args:
            base_key: a string to group entries under, defaults to DEFAULT_BASE_KEY,
                      can be a callable taking the request and cache instance as arguments
            expires_ttl: length in seconds entries should live in the cache, EXPIRES_TTL is default
            keys: a list of CacheKeys to use for key creation upon entry placement, path is default
            auth: a callable to check authorization of the request; takes the request, cache
                  instance, and base_key as arguments.
        """
        super().__init__()
        self.default_base_key = base_key or self.default_base_key
        self.keys = keys or (
            CacheKeys.path,
            CacheKeys.method,
        )
        self.default_expires_ttl = expires_ttl or self.default_expires_ttl
        self.auth = auth

    def __call__(self, func):
        """Decorator magic call to make the handler cached"""
        if not settings.CACHE_ENABLED:
            return func

        async def cached_function(*args):
            request = self.get_request_from_args(args)
            bk = self.default_base_key
            if callable(self.default_base_key):
                bk = await self.default_base_key(request, self)
            if self.auth:
                await self.auth(request, self, bk)
            key = self.make_key(request)
            # Check cache
            response = await self.make_response(key, bk)
            if response is None:
                # Cache miss, create new entry
                response = await self.make_entry(key, bk, func, args, self.default_expires_ttl)
            return response

        return cached_function

    def get_request_from_args(self, args):
        """Finds the request object from list of args"""
        for arg in args:
            if isinstance(arg, Request):
                return arg

    async def make_response(self, key, base_key):
        """Tries to find the cached entry and turn it into a proper response"""
        entry = await self.get(key, base_key)
        if not entry:
            return None
        entry = json.loads(entry)
        response_type = entry.pop("type", None)
        if not response_type or response_type not in self.RESPONSE_TYPES:
            # Bad entry, delete from cache
            self.delete(key, base_key)
            return None
        response = self.RESPONSE_TYPES[response_type](**entry)
        response.headers.update({"X-PULP-CACHE": "HIT"})
        return response

    async def make_entry(self, key, base_key, handler, args, expires=86400):
        """Gets the response for the request and try to turn it into a cacheable entry"""
        try:
            response = await handler(*args)
        except (HTTPSuccessful, HTTPFound) as e:
            response = e

        entry = {"headers": dict(response.headers), "status": response.status}
        response.headers.update({"X-PULP-CACHE": "MISS"})
        if isinstance(response, FileResponse):
            entry["path"] = str(response._path)
            entry["type"] = "FileResponse"
        elif isinstance(response, (Response, HTTPSuccessful)):
            body = response.body
            entry["text"] = getattr(body, "_value", body).decode("utf-8")
            entry["type"] = "Response"
        elif isinstance(response, HTTPFound):
            entry["location"] = str(response.location)
            entry["type"] = "Redirect"
        else:
            # We don't cache StreamResponses or errors
            return response

        # TODO look into smaller format, maybe some compression on the text
        await self.set(key, json.dumps(entry), expires, base_key=base_key)
        return response

    def make_key(self, request):
        """Makes the key based off the request"""
        # Might potentially have to make this async if keys require async data from request
        all_keys = {
            CacheKeys.path: request.match_info["path"],
            CacheKeys.method: request.method,
            CacheKeys.host: request.url.host,
        }
        key = ":".join(all_keys[k] for k in self.keys)
        return key


# TODO Add Cache object for non async redis connection/ aka Django requests



"""Test the status page."""
import warnings
import unittest

from django.test import override_settings
from jsonschema import validate
from pulp_smash import api, cli, config, utils
from pulp_smash.pulp3.constants import STATUS_PATH
from requests.exceptions import HTTPError

from pulpcore.tests.functional.api.utils import get_redis_status
from pulpcore.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

STATUS = {
    "$schema": "http://json-schema.org/schema#",
    "title": "Pulp 3 status API schema",
    "description": ("Derived from Pulp's actual behaviour and various Pulp issues."),
    "type": "object",
    "properties": {
        "database_connection": {
            "type": "object",
            "properties": {"connected": {"type": "boolean"}},
        },
        "redis_connection": {"type": "object", "properties": {"connected": {"type": "boolean"}}},
        "missing_workers": {"type": "array", "items": {"type": "object"}},
        "online_workers": {"type": "array", "items": {"type": "object"}},
        "versions": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {"component": {"type": "string"}, "version": {"type": "string"}},
            },
        },
        "storage": {
            "type": "object",
            "properties": {
                "total": {"type": "integer"},
                "used": {"type": "integer"},
                "free": {"type": "integer"},
            },
        },
    },
}


class StatusTestCase(unittest.TestCase):
    """Tests related to the status page.

    This test explores the following issues:

    * `Pulp #2804 <https://pulp.plan.io/issues/2804>`_
    * `Pulp #2867 <https://pulp.plan.io/issues/2867>`_
    * `Pulp #3544 <https://pulp.plan.io/issues/3544>`_
    * `Pulp Smash #755 <https://github.com/pulp/pulp-smash/issues/755>`_
    """

    def setUp(self):
        """Make an API client."""
        self.client = api.Client(config.get_config(), api.json_handler)
        self.status_response = STATUS
        cli_client = cli.Client(config.get_config())
        self.storage = utils.get_pulp_setting(cli_client, "DEFAULT_FILE_STORAGE")

        if self.storage != "pulpcore.app.models.storage.FileSystem":
            self.status_response["properties"].pop("storage", None)

        self.is_redis_connected = get_redis_status()

    def test_get_authenticated(self):
        """GET the status path with valid credentials.

        Verify the response with :meth:`verify_get_response`.
        """
        self.verify_get_response(self.client.get(STATUS_PATH))

    def test_get_unauthenticated(self):
        """GET the status path with no credentials.

        Verify the response with :meth:`verify_get_response`.
        """
        del self.client.request_kwargs["auth"]
        self.verify_get_response(self.client.get(STATUS_PATH))

    def test_post_authenticated(self):
        """POST the status path with valid credentials.

        Assert an error is returned.
        """
        with self.assertRaises(HTTPError):
            self.client.post(STATUS_PATH)

    def verify_get_response(self, status):
        """Verify the response to an HTTP GET call.

        Verify that several attributes and have the correct type or value.
        """
        validate(status, self.status_response)
        self.assertTrue(status["database_connection"]["connected"])
        self.assertNotEqual(status["online_workers"], [])
        self.assertNotEqual(status["versions"], [])
        if self.storage == "pulpcore.app.models.storage.FileSystem":
            self.assertIsNotNone(status["storage"])
        else:
            self.assertIsNone(status["storage"])

        self.assertIsNotNone(status["redis_connection"])
        if self.is_redis_connected:
            self.assertTrue(status["redis_connection"]["connected"])
        else:
            warnings.warn("Could not connect to the Redis server")

    @override_settings(CACHE_ENABLED=False)
    def verify_get_response_without_redis(self, status):
        """Verify the response to an HTTP GET call when Redis is not used.

        Verify that redis_connection is null
        """
        validate(status, self.status_response)
        self.assertIsNone(status["redis_connection"])



import os
import shutil

from django.conf import settings


class _WorkingDir:
    """
    A base class for temporary working directories.

    TODO: This is very similar to functionality already in the stdlib, why keep our own?

    https://docs.python.org/3/library/tempfile.html#tempfile.TemporaryDirectory
    """

    # Directory permissions.
    MODE = 0o700

    def __init__(self, path):
        """
        Args:
            path (str): The absolute directory path to use.
        """
        self._path = path

    @property
    def path(self):
        """
        The absolute path to the directory.

        Returns:
            str: The absolute directory path.
        """
        return self._path

    def create(self):
        """
        Create the directory.
        """
        os.makedirs(self.path, mode=self.MODE)

    def delete(self):
        """
        Delete the directory.

        On permission denied - an attempt is made to fix the
        permissions on the tree and the delete is retried.
        """
        try:
            self._delete()
        except PermissionError:
            self._set_permissions()
            self._delete()

    def _delete(self):
        """
        Helper method for delete
        """
        try:
            shutil.rmtree(self.path, ignore_errors=True)
        except FileNotFoundError:
            pass

    def _set_permissions(self):
        """
        Set appropriate permissions on the directory tree.
        """
        for path in os.walk(self.path):
            os.chmod(path[0], mode=self.MODE)

    def __str__(self):
        return self.path

    def __enter__(self):
        """
        Create the directory and set the CWD to the path.

        Returns: self

        Raises:
            OSError: On failure.
        """
        self.create()
        self._prev_dir = os.getcwd()
        os.chdir(self._path)
        return self

    def __exit__(self, *unused):
        """
        Delete the directory (tree) and restore the original CWD.
        """
        os.chdir(self._prev_dir)
        self.delete()


def get_worker_path(hostname):
    """
    Get the root directory path for a worker by hostname.

    Format: <root>/<worker-hostname>

    Args:
        hostname (str): The worker hostname.

    Returns:
        str: The absolute path to a worker's root directory.
    """
    return os.path.join(settings.WORKING_DIRECTORY, hostname)


class WorkerDirectory(_WorkingDir):
    """
    The directory associated with a pulpcore-worker.

    Path format: <root>/<worker-hostname>
    """

    def __init__(self, hostname):
        """
        Args:
            hostname (str): The worker hostname.
        """
        self._path = get_worker_path(hostname)

    def create(self):
        """
        Create the directory.

        The directory is deleted and recreated when already exists.
        Only one of these should ever be held at a time for any individual worker.
        """
        try:
            super().create()
        except FileExistsError:
            self.delete()
            super().create()



from gettext import gettext as _

import asyncio
import importlib
import json
import logging
import os
import random
import select
import signal
import socket
import sys
import traceback
from contextlib import suppress
from datetime import timedelta
from multiprocessing import Process
from tempfile import TemporaryDirectory

import django

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "pulpcore.app.settings")
django.setup()

from django.conf import settings  # noqa: E402: module level not at top of file
from django.db import connection  # noqa: E402: module level not at top of file
from django.utils import timezone  # noqa: E402: module level not at top of file
from django_currentuser.middleware import (  # noqa: E402: module level not at top of file
    _set_current_user,
)
from django_guid import set_guid  # noqa: E402: module level not at top of file

from pulpcore.app.models import Worker, Task  # noqa: E402: module level not at top of file

from pulpcore.app.role_util import (  # noqa: E402: module level not at top of file
    get_users_with_perms,
)

from pulpcore.constants import (  # noqa: E402: module level not at top of file
    TASK_STATES,
    TASK_INCOMPLETE_STATES,
)

from pulpcore.exceptions import AdvisoryLockError  # noqa: E402: module level not at top of file
from pulpcore.tasking.storage import WorkerDirectory  # noqa: E402: module level not at top of file
from pulpcore.tasking.util import _delete_incomplete_resources  # noqa: E402


_logger = logging.getLogger(__name__)
random.seed()

TASK_GRACE_INTERVAL = 3
WORKER_CLEANUP_INTERVAL = 100


def handle_worker_heartbeat(worker_name):
    """
    This is a generic function for updating worker heartbeat records.

    Existing Worker objects are searched for one to update. If an existing one is found, it is
    updated. Otherwise a new Worker entry is created. Logging at the info level is also done.

    Args:
        worker_name (str): The hostname of the worker
    """
    worker, created = Worker.objects.get_or_create(name=worker_name)

    if created:
        _logger.info(_("New worker '{name}' discovered").format(name=worker_name))
    elif worker.online is False:
        _logger.info(_("Worker '{name}' is back online.").format(name=worker_name))

    worker.save_heartbeat()

    msg = _("Worker heartbeat from '{name}' at time {timestamp}").format(
        timestamp=worker.last_heartbeat, name=worker_name
    )

    _logger.debug(msg)

    return worker


class NewPulpWorker:
    def __init__(self):
        self.shutdown_requested = False
        self.name = f"{os.getpid()}@{socket.getfqdn()}"
        self.heartbeat_period = settings.WORKER_TTL / 3
        self.cursor = connection.cursor()
        self.worker = handle_worker_heartbeat(self.name)
        self.task_grace_timeout = 0
        self.worker_cleanup_countdown = random.randint(
            WORKER_CLEANUP_INTERVAL / 10, WORKER_CLEANUP_INTERVAL
        )

        # Add a file descriptor to trigger select on signals
        self.sentinel, sentinel_w = os.pipe()
        os.set_blocking(self.sentinel, False)
        os.set_blocking(sentinel_w, False)
        signal.set_wakeup_fd(sentinel_w)

    def _signal_handler(self, thesignal, frame):
        # Reset signal handlers to default
        # If you kill the process a second time it's not graceful anymore.
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)

        _logger.info(_("Worker %s was requested to shut down."), self.name)

        self.shutdown_requested = True

    def shutdown(self):
        self.worker.delete()
        _logger.info(_("Worker %s was shut down."), self.name)

    def worker_cleanup(self):
        qs = Worker.objects.missing_workers(age=timedelta(days=7))
        if qs:
            for worker in qs:
                _logger.info(_("Clean missing worker %s."), worker.name)
            qs.delete()

    def beat(self):
        if self.worker.last_heartbeat < timezone.now() - timedelta(seconds=self.heartbeat_period):
            self.worker = handle_worker_heartbeat(self.name)
            if self.shutdown_requested:
                self.task_grace_timeout -= 1
            self.worker_cleanup_countdown -= 1
            if self.worker_cleanup_countdown <= 0:
                self.worker_cleanup_countdown = WORKER_CLEANUP_INTERVAL
                self.worker_cleanup()

    def notify_workers(self):
        self.cursor.execute("NOTIFY pulp_worker_wakeup")

    def cancel_abandoned_task(self, task, final_state, reason=None):
        """Cancel and clean up an abandoned task.

        This function must only be called while holding the lock for that task. It is a no-op if
        the task is neither in "running" nor "canceling" state.

        Return ``True`` if the task was actually canceled, ``False`` otherwise.
        """
        # A task is considered abandoned when in running state, but no worker holds its lock
        Task.objects.filter(pk=task.pk, state=TASK_STATES.RUNNING).update(
            state=TASK_STATES.CANCELING
        )
        task.refresh_from_db()
        if task.state == TASK_STATES.CANCELING:
            if reason:
                _logger.info(
                    _("Cleaning up task %s and marking as %s. Reason: %s"),
                    task.pk,
                    final_state,
                    reason,
                )
            else:
                _logger.info(_("Cleaning up task %s and marking as %s."), task.pk, final_state)
            _delete_incomplete_resources(task)
            if task.reserved_resources_record:
                self.notify_workers()
            task_data = {
                "state": final_state,
                "finished_at": timezone.now(),
            }
            if reason:
                task_data["error"] = {"reason": reason}
            Task.objects.filter(pk=task.pk, state=TASK_STATES.CANCELING).update(**task_data)
            return True
        return False

    def iter_tasks(self):
        """Iterate over ready tasks and yield each task while holding the lock."""

        while not self.shutdown_requested:
            taken_exclusive_resources = set()
            taken_shared_resources = set()
            # When batching this query, be sure to use "pulp_created" as a cursor
            for task in Task.objects.filter(state__in=TASK_INCOMPLETE_STATES).order_by(
                "pulp_created"
            ):
                reserved_resources_record = task.reserved_resources_record or []
                exclusive_resources = [
                    resource
                    for resource in reserved_resources_record
                    if not resource.startswith("shared:")
                ]
                shared_resources = [
                    resource[7:]
                    for resource in reserved_resources_record
                    if resource.startswith("shared:") and resource[7:] not in exclusive_resources
                ]
                with suppress(AdvisoryLockError), task:
                    # This code will only be called if we acquired the lock successfully
                    # The lock will be automatically be released at the end of the block
                    # Check if someone else changed the task before we got the lock
                    task.refresh_from_db()
                    if task.state == TASK_STATES.CANCELING and task.worker is None:
                        # No worker picked this task up before being canceled
                        if self.cancel_abandoned_task(task, TASK_STATES.CANCELED):
                            # Continue looking for the next task
                            # without considering this tasks resources
                            # as we just released them
                            continue
                    if task.state in [TASK_STATES.RUNNING, TASK_STATES.CANCELING]:
                        # A running task without a lock must be abandoned
                        if self.cancel_abandoned_task(
                            task, TASK_STATES.FAILED, "Worker has gone missing."
                        ):
                            # Continue looking for the next task
                            # without considering this tasks resources
                            # as we just released them
                            continue
                    # This statement is using lazy evaluation
                    if (
                        task.state == TASK_STATES.WAITING
                        # No exclusive resource taken?
                        and not any(
                            resource in taken_exclusive_resources
                            or resource in taken_shared_resources
                            for resource in exclusive_resources
                        )
                        # No shared resource exclusively taken?
                        and not any(
                            resource in taken_exclusive_resources for resource in shared_resources
                        )
                    ):
                        yield task
                        # Start from the top of the Task list
                        break

                # Record the resources of the pending task we didn't get
                taken_exclusive_resources.update(exclusive_resources)
                taken_shared_resources.update(shared_resources)
            else:
                # If we got here, there is nothing to do
                break

    def sleep(self):
        """Wait for signals on the wakeup channel while heart beating."""

        _logger.debug(_("Worker %s entering sleep state."), self.name)
        wakeup = False
        while not self.shutdown_requested:
            # Handle all notifications before sleeping in `select`
            while connection.connection.notifies:
                item = connection.connection.notifies.pop(0)
                if item.channel == "pulp_worker_wakeup":
                    _logger.debug(_("Worker %s received wakeup call."), self.name)
                    wakeup = True
                # ignore all other notifications
            if wakeup:
                break

            r, w, x = select.select(
                [self.sentinel, connection.connection], [], [], self.heartbeat_period
            )
            self.beat()
            if connection.connection in r:
                connection.connection.poll()
            if self.sentinel in r:
                os.read(self.sentinel, 256)

    def supervise_task(self, task):
        """Call and supervise the task process while heart beating.

        This function must only be called while holding the lock for that task."""

        self.task_grace_timeout = TASK_GRACE_INTERVAL
        task.worker = self.worker
        task.save(update_fields=["worker"])
        cancel_state = None
        cancel_reason = None
        with TemporaryDirectory(dir=".") as task_working_dir_rel_path:
            task_process = Process(target=_perform_task, args=(task.pk, task_working_dir_rel_path))
            task_process.start()
            while True:
                # Handle all notifications before sleeping in `select`
                while connection.connection.notifies:
                    item = connection.connection.notifies.pop(0)
                    if item.channel == "pulp_worker_cancel" and item.payload == str(task.pk):
                        _logger.info(_("Received signal to cancel current task %s."), task.pk)
                        os.kill(task_process.pid, signal.SIGUSR1)
                        cancel_state = TASK_STATES.CANCELED
                    # ignore all other notifications
                if cancel_state:
                    break

                r, w, x = select.select(
                    [self.sentinel, connection.connection, task_process.sentinel],
                    [],
                    [],
                    self.heartbeat_period,
                )
                self.beat()
                if connection.connection in r:
                    connection.connection.poll()
                if task_process.sentinel in r:
                    if not task_process.is_alive():
                        break
                if self.sentinel in r:
                    os.read(self.sentinel, 256)
                if self.shutdown_requested:
                    if self.task_grace_timeout > 0:
                        _logger.info(
                            _("Worker shutdown requested, waiting for task %s to finish."), task.pk
                        )
                    else:
                        _logger.info(_("Aborting current task %s due to worker shutdown."), task.pk)
                        os.kill(task_process.pid, signal.SIGUSR1)
                        cancel_state = TASK_STATES.FAILED
                        cancel_reason = "Aborted during worker shutdown."
                        break
            task_process.join()
            if not cancel_state and task_process.exitcode != 0:
                _logger.warn(
                    _("Task process for %s exited with non zero exitcode %i."),
                    task.pk,
                    task_process.exitcode,
                )
                cancel_state = TASK_STATES.FAILED
                if task_process.exitcode < 0:
                    cancel_reason = "Killed by signal {sig_num}.".format(
                        sig_num=-task_process.exitcode
                    )
                else:
                    cancel_reason = "Task process died unexpectedly with exitcode {code}.".format(
                        code=task_process.exitcode
                    )
            if cancel_state:
                self.cancel_abandoned_task(task, cancel_state, cancel_reason)
        if task.reserved_resources_record:
            self.notify_workers()

    def run_forever(self):
        with WorkerDirectory(self.name):
            signal.signal(signal.SIGINT, self._signal_handler)
            signal.signal(signal.SIGTERM, self._signal_handler)
            # Subscribe to pgsql channels
            self.cursor.execute("LISTEN pulp_worker_wakeup")
            self.cursor.execute("LISTEN pulp_worker_cancel")
            while not self.shutdown_requested:
                for task in self.iter_tasks():
                    self.supervise_task(task)
                if not self.shutdown_requested:
                    self.sleep()
            self.cursor.execute("UNLISTEN pulp_worker_cancel")
            self.cursor.execute("UNLISTEN pulp_worker_wakeup")
            self.shutdown()


def child_signal_handler(sig, frame):
    # Reset signal handlers to default
    # If you kill the process a second time it's not graceful anymore.
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    signal.signal(signal.SIGTERM, signal.SIG_DFL)
    signal.signal(signal.SIGUSR1, signal.SIG_DFL)

    if sig == signal.SIGUSR1:
        sys.exit()


def _perform_task(task_pk, task_working_dir_rel_path):
    """Setup the environment to handle a task and execute it.
    This must be called as a subprocess, while the parent holds the advisory lock."""
    signal.signal(signal.SIGINT, child_signal_handler)
    signal.signal(signal.SIGTERM, child_signal_handler)
    signal.signal(signal.SIGUSR1, child_signal_handler)
    # All processes need to create their own postgres connection
    connection.connection = None
    task = Task.objects.get(pk=task_pk)
    task.set_running()
    # Store the task id in the environment for `Task.current()`.
    os.environ["PULP_TASK_ID"] = str(task.pk)
    user = get_users_with_perms(task, with_group_users=False).first()
    _set_current_user(user)
    set_guid(task.logging_cid)
    try:
        _logger.info(_("Starting task %s"), task.pk)

        # Execute task
        module_name, function_name = task.name.rsplit(".", 1)
        module = importlib.import_module(module_name)
        func = getattr(module, function_name)
        args = json.loads(task.args) or ()
        kwargs = json.loads(task.kwargs) or {}
        os.chdir(task_working_dir_rel_path)
        result = func(*args, **kwargs)
        if asyncio.iscoroutine(result):
            _logger.debug(_("Task is coroutine %s"), task.pk)
            loop = asyncio.get_event_loop()
            loop.run_until_complete(result)

    except Exception:
        exc_type, exc, tb = sys.exc_info()
        task.set_failed(exc, tb)
        _logger.info(_("Task %s failed (%s)"), task.pk, exc)
        _logger.info("\n".join(traceback.format_list(traceback.extract_tb(tb))))
    else:
        task.set_completed()
        _logger.info(_("Task completed %s"), task.pk)
    os.environ.pop("PULP_TASK_ID")



# Generated by Django 2.2.12 on 2020-06-09 12:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0032_export_to_chunks'),
    ]

    operations = [
        migrations.AlterField(
            model_name='remoteartifact',
            name='size',
            field=models.BigIntegerField(null=True),
        ),
    ]



# Generated by Django 3.2.7 on 2021-09-22 12:55

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0078_grouprole_role_userrole'),
    ]

    operations = [
        migrations.RenameField(
            model_name='accesspolicy',
            old_name='permissions_assignment',
            new_name='creation_hooks',
        ),
    ]



# Generated by Django 3.2.6 on 2021-09-29 14:00

from urllib.parse import urlparse, urlunparse

from django.db import migrations


def move_remote_url_credentials(apps, schema_editor):
    Remote = apps.get_model("core", "Remote")

    for remote in Remote.objects.filter(url__contains="@").iterator():
        url = urlparse(remote.url)

        if not remote.username:
            remote.username = url.username
        if not remote.password:
            remote.password = url.password

        _, url_split = url.netloc.rsplit("@", maxsplit=1)
        remote.url = urlunparse(url._replace(netloc=url_split))
        remote.save()


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0076_remove_reserved_resource'),
    ]

    operations = [
        migrations.RunPython(
            code=move_remote_url_credentials,
            reverse_code=migrations.RunPython.noop,
            elidable=True,
        )
    ]



# Generated by Django 2.2.12 on 2020-06-23 13:57

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0033_increase_remote_artifact_size_field'),
    ]

    operations = [
        migrations.CreateModel(
            name='GroupProgressReport',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('message', models.TextField()),
                ('code', models.CharField(max_length=36)),
                ('total', models.IntegerField(default=0)),
                ('done', models.IntegerField(default=0)),
                ('suffix', models.TextField(null=True)),
                ('task_group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='group_progress_reports', to='core.TaskGroup')),
            ],
            options={
                'unique_together': {('code', 'task_group')},
            },
        ),
    ]



"""
Django settings for the Pulp Platform application

Never import this module directly, instead `from django.conf import settings`, see
https://docs.djangoproject.com/en/1.11/topics/settings/#using-settings-in-python-code

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.11/ref/settings/
"""

import sys

from contextlib import suppress
from gettext import gettext as _
from importlib import import_module
from logging import getLogger
from pathlib import Path
from pkg_resources import iter_entry_points

from cryptography.fernet import Fernet
from django.core.exceptions import ImproperlyConfigured
from django.db import connection

from pulpcore import constants

# Build paths inside the project like this: BASE_DIR / ...
BASE_DIR = Path(__file__).absolute().parent

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.11/howto/deployment/checklist/

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = False

ALLOWED_HOSTS = ["*"]

DEPLOY_ROOT = Path("/var/lib/pulp")
MEDIA_ROOT = str(DEPLOY_ROOT / "media")  # Django 3.1 adds support for pathlib.Path

ADMIN_SITE_URL = "admin/"

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.11/howto/static-files/

STATIC_URL = "/assets/"
STATIC_ROOT = DEPLOY_ROOT / STATIC_URL.strip("/")

DEFAULT_FILE_STORAGE = "pulpcore.app.models.storage.FileSystem"

WORKING_DIRECTORY = DEPLOY_ROOT / "tmp"
FILE_UPLOAD_TEMP_DIR = WORKING_DIRECTORY

CHUNKED_UPLOAD_DIR = "upload"

# List of upload handler classes to be applied in order.
FILE_UPLOAD_HANDLERS = ("pulpcore.app.files.HashingFileUploadHandler",)

SECRET_KEY = True

# Key used to encrypt fields in the database
DB_ENCRYPTION_KEY = "/etc/pulp/certs/database_fields.symmetric.key"

# Application definition

INSTALLED_APPS = [
    # django stuff
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "import_export",
    # third-party
    "django_filters",
    "django_guid",
    "drf_spectacular",
    "guardian",
    "rest_framework",
    # pulp core app
    "pulpcore.app",
]

# Enumerate the installed Pulp plugins during the loading process for use in the status API
INSTALLED_PULP_PLUGINS = []

for entry_point in iter_entry_points("pulpcore.plugin"):
    plugin_app_config = entry_point.load()
    INSTALLED_PULP_PLUGINS.append(entry_point.module_name)
    INSTALLED_APPS.append(plugin_app_config)

# Optional apps that help with development, or augment Pulp in some non-critical way
OPTIONAL_APPS = [
    "crispy_forms",
    "django_extensions",
    "storages",
]

for app in OPTIONAL_APPS:
    # only import if app is installed
    with suppress(ImportError):
        import_module(app)
        INSTALLED_APPS.append(app)

MIDDLEWARE = [
    "django_guid.middleware.guid_middleware",
    "django.middleware.security.SecurityMiddleware",
    "whitenoise.middleware.WhiteNoiseMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django_currentuser.middleware.ThreadLocalUserMiddleware",
]

AUTHENTICATION_BACKENDS = [
    "django.contrib.auth.backends.ModelBackend",
    "guardian.backends.ObjectPermissionBackend",
    "pulpcore.backends.ObjectRolePermissionBackend",
]

# Disable django guardian anonymous user
# https://django-guardian.readthedocs.io/en/stable/configuration.html#anonymous-user-name
ANONYMOUS_USER_NAME = None

ROOT_URLCONF = "pulpcore.app.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [BASE_DIR / "templates"],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "pulpcore.app.wsgi.application"

REST_FRAMEWORK = {
    "URL_FIELD_NAME": "pulp_href",
    "DEFAULT_FILTER_BACKENDS": ("django_filters.rest_framework.DjangoFilterBackend",),
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.LimitOffsetPagination",
    "PAGE_SIZE": 100,
    "DEFAULT_PERMISSION_CLASSES": ("pulpcore.plugin.access_policy.AccessPolicyFromDB",),
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "rest_framework.authentication.SessionAuthentication",
        "rest_framework.authentication.BasicAuthentication",
    ),
    "UPLOADED_FILES_USE_URL": False,
    "DEFAULT_VERSIONING_CLASS": "rest_framework.versioning.URLPathVersioning",
    "DEFAULT_SCHEMA_CLASS": "pulpcore.openapi.PulpAutoSchema",
}

# Password validation
# https://docs.djangoproject.com/en/dev/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {"NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"},
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]


# Internationalization
# https://docs.djangoproject.com/en/1.11/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = "USE_I18N", True

USE_L10N = True

USE_TZ = True


# A set of default settings to use if the configuration file in
# /etc/pulp/ is missing or if it does not have values for every setting

# https://docs.djangoproject.com/en/1.11/ref/settings/#databases
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": "pulp",
        "USER": "pulp",
        "CONN_MAX_AGE": 0,
    },
}

# Redis default config
REDIS_URL = None
REDIS_HOST = "localhost"
REDIS_PORT = 6379
REDIS_DB = 0
REDIS_PASSWORD = None
REDIS_SSL = False
REDIS_SSL_CA_CERTS = None

# https://docs.djangoproject.com/en/1.11/ref/settings/#logging and
# https://docs.python.org/3/library/logging.config.html
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "simple": {"format": "pulp [%(correlation_id)s]: %(name)s:%(levelname)s: %(message)s"}
    },
    "filters": {"correlation_id": {"()": "django_guid.log_filters.CorrelationId"}},
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "simple",
            "filters": ["correlation_id"],
        }
    },
    "loggers": {
        "": {
            # The root logger
            "handlers": ["console"],
            "level": "INFO",
            "filters": ["correlation_id"],
        },
        "django_guid": {
            "handlers": ["console"],
            "level": "WARNING",
            "propagate": False,
        },
    },
}

DRF_ACCESS_POLICY = {"reusable_conditions": "pulpcore.app.global_access_conditions"}

CONTENT_PATH_PREFIX = "/pulp/content/"
CONTENT_APP_TTL = 30

WORKER_TTL = 30

# how long to protect orphan content in minutes
ORPHAN_PROTECTION_TIME = 24 * 60

REMOTE_USER_ENVIRON_NAME = "REMOTE_USER"

ALLOWED_IMPORT_PATHS = []

ALLOWED_EXPORT_PATHS = []

PROFILE_STAGES_API = False

# https://docs.pulpproject.org/pulpcore/configuration/settings.html#pulp-cache
CACHE_ENABLED = True
CACHE_SETTINGS = {
    "EXPIRES_TTL": 600,  # 10 minutes
}

SPECTACULAR_SETTINGS = {
    "SERVE_URLCONF": ROOT_URLCONF,
    "DEFAULT_GENERATOR_CLASS": "pulpcore.openapi.PulpSchemaGenerator",
    "DEFAULT_SCHEMA_CLASS": "pulpcore.openapi.PulpAutoSchema",
    "ENUM_ADD_EXPLICIT_BLANK_NULL_CHOICE": False,
    "COMPONENT_SPLIT_REQUEST": True,
    "COMPONENT_NO_READ_ONLY_REQUIRED": True,
    "GENERIC_ADDITIONAL_PROPERTIES": None,
    "DISABLE_ERRORS_AND_WARNINGS": not DEBUG,
    "TITLE": "Pulp 3 API",
    "DESCRIPTION": "Fetch, Upload, Organize, and Distribute Software Packages",
    "VERSION": "v3",
    "CONTACT": {
        "name": "Pulp Team",
        "email": "pulp-list@redhat.com",
        "url": "https://pulpproject.org",
    },
    "LICENSE": {
        "name": "GPLv2+",
        "url": "https://raw.githubusercontent.com/pulp/pulpcore/master/LICENSE",
    },
}

# What kinds of checksums is this pulp-instance _allowed to use_ ?
# NOTE : "sha256"" IS REQUIRED - Pulp will fail to start if it is not found in this set
# NOTE: specifying checksums that are not listed under ALL_KNOWN_CONTENT_CHECKSUMS will fail
#       at startup
ALLOWED_CONTENT_CHECKSUMS = ["sha224", "sha256", "sha384", "sha512"]

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

# HERE STARTS DYNACONF EXTENSION LOAD (Keep at the very bottom of settings.py)
# Read more at https://dynaconf.readthedocs.io/en/latest/guides/django.html
import dynaconf  # noqa

settings = dynaconf.DjangoDynaconf(
    __name__,
    GLOBAL_ENV_FOR_DYNACONF="PULP",
    ENV_SWITCHER_FOR_DYNACONF="PULP_ENV",
    PRELOAD_FOR_DYNACONF=[
        "{}.app.settings".format(plugin_name) for plugin_name in INSTALLED_PULP_PLUGINS
    ],
    ENVVAR_FOR_DYNACONF="PULP_SETTINGS",
    load_dotenv=False,
)
# HERE ENDS DYNACONF EXTENSION LOAD (No more code below this line)

_logger = getLogger(__name__)

try:
    CONTENT_ORIGIN
except NameError:
    raise ImproperlyConfigured(
        _(
            "CONTENT_ORIGIN is a required setting but it was not configured. This may be caused "
            "by invalid read permissions of the settings file. Note that CONTENT_ORIGIN is set by "
            "the installer automatically."
        )
    )

if not (
    Path(sys.argv[0]).name == "pytest"
    or Path(sys.argv[0]).name == "sphinx-build"
    or (len(sys.argv) >= 2 and sys.argv[1] == "collectstatic")
):
    try:
        with open(DB_ENCRYPTION_KEY, "rb") as key_file:
            Fernet(key_file.read())
    except Exception as ex:
        raise ImproperlyConfigured(
            _("Could not load DB_ENCRYPTION_KEY file '{file}': {err}").format(
                file=DB_ENCRYPTION_KEY, err=ex
            )
        )

# Check legality of ALLOWED_CONTENT_CHECKSUMS post-dynaconf-load, in case it has been overridden
# in a site-specific location (eg, in /etc/pulp/settings.py)
if "sha256" not in ALLOWED_CONTENT_CHECKSUMS:
    raise ImproperlyConfigured(
        _(
            "ALLOWED_CONTENT_CHECKSUMS MUST contain 'sha256' - Pulp's content-storage-addressing "
            "relies on sha256 to identify entities."
        )
    )

unknown_algs = set(ALLOWED_CONTENT_CHECKSUMS).difference(constants.ALL_KNOWN_CONTENT_CHECKSUMS)
if unknown_algs:
    raise ImproperlyConfigured(
        _(
            "ALLOWED_CONTENT_CHECKSUMS may only contain algorithms known to pulp - see "
            "constants.ALL_KNOWN_CONTENT_CHECKSUMS for the allowed list. Unknown algorithms "
            "provided: {}".format(unknown_algs)
        )
    )

FORBIDDEN_CHECKSUMS = set(constants.ALL_KNOWN_CONTENT_CHECKSUMS).difference(
    ALLOWED_CONTENT_CHECKSUMS
)

_SKIPPED_COMMANDS_FOR_CONTENT_CHECKS = ["handle-artifact-checksums", "migrate", "collectstatic"]

if not (len(sys.argv) >= 2 and sys.argv[1] in _SKIPPED_COMMANDS_FOR_CONTENT_CHECKS):
    try:
        with connection.cursor() as cursor:
            for checksum in ALLOWED_CONTENT_CHECKSUMS:
                # can't import Artifact here so use a direct db connection
                cursor.execute(f"SELECT count(pulp_id) FROM core_artifact WHERE {checksum} IS NULL")
                row = cursor.fetchone()
                if row[0] > 0:
                    raise ImproperlyConfigured(
                        _(
                            "There have been identified artifacts missing checksum '{}'. "
                            "Run 'pulpcore-manager handle-artifact-checksums' first to populate "
                            "missing artifact checksums."
                        ).format(checksum)
                    )
            for checksum in FORBIDDEN_CHECKSUMS:
                # can't import Artifact here so use a direct db connection
                cursor.execute(
                    f"SELECT count(pulp_id) FROM core_artifact WHERE {checksum} IS NOT NULL"
                )
                row = cursor.fetchone()
                if row[0] > 0:
                    raise ImproperlyConfigured(
                        _(
                            "There have been identified artifacts with forbidden checksum '{}'. "
                            "Run 'pulpcore-manager handle-artifact-checksums' first to unset "
                            "forbidden checksums."
                        ).format(checksum)
                    )

            # warn if there are remote artifacts with checksums but no allowed checksums
            cond = " AND ".join([f"{c} IS NULL" for c in constants.ALL_KNOWN_CONTENT_CHECKSUMS])
            no_checksum_query = f"SELECT pulp_id FROM core_remoteartifact WHERE {cond}"
            cond = " AND ".join([f"{c} IS NULL" for c in ALLOWED_CONTENT_CHECKSUMS])
            cursor.execute(
                f"SELECT count(pulp_id) FROM core_remoteartifact WHERE {cond} AND "
                f"pulp_id NOT IN ({no_checksum_query})"
            )
            row = cursor.fetchone()
            if row[0] > 0:
                _logger.warn(
                    _(
                        "Warning: detected remote content without allowed checksums. "
                        "Run 'pulpcore-manager handle-artifact-checksums --report' to "
                        "view this content."
                    )
                )

    except ImproperlyConfigured as e:
        raise e
    except Exception:
        # our check could fail if the table hasn't been created yet or we can't get a db connection
        pass
    finally:
        connection.close()



# Generated by Django 2.2.6 on 2019-10-28 18:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0012_auto_20191104_2000'),
    ]

    operations = [
        migrations.AddField(
            model_name='repository',
            name='pulp_type',
            field=models.TextField(default=None),
        ),
    ]



# Generated by Django 2.2.9 on 2020-01-21 13:05

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0019_add_signing_service_model'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='publishedartifact',
            unique_together={('publication', 'relative_path')},
        ),
    ]



from setuptools import find_packages, setup

with open("README.md") as f:
    long_description = f.read()

with open("requirements.txt") as requirements:
    requirements = requirements.readlines()

setup(
    name="pulpcore",
    version="3.17.8",
    description="Pulp Django Application and Related Modules",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GPLv2+",
    packages=find_packages(exclude=["test"]),
    author="Pulp Team",
    author_email="pulp-list@redhat.com",
    url="https://pulpproject.org",
    python_requires=">=3.8",
    install_requires=requirements,
    extras_require={
        "s3": ["django-storages[boto3]"],
        "azure": ["django-storages[azure]>=1.12.2"],
        "prometheus": ["django-prometheus"],
    },
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    scripts=["bin/pulp-content"],
    entry_points={
        "console_scripts": [
            "pulpcore-manager = pulpcore.app.manage:manage",
            "pulpcore-worker = pulpcore.tasking.entrypoint:worker",
        ]
    },
)



"""Tests for Pulp 3's authentication API.

For more information, see the documentation on `Authentication
<https://docs.pulpproject.org/restapi.html#section/Authentication>`_.
"""
import unittest

from pulp_smash import api, config, utils
from pulp_smash.pulp3.constants import ARTIFACTS_PATH
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError

from pulpcore.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


class AuthTestCase(unittest.TestCase):
    """Test Pulp3 Authentication."""

    def setUp(self):
        """Set common variable."""
        self.cfg = config.get_config()

    def test_base_auth_success(self):
        """Perform HTTP basic authentication with valid credentials.

        Assert that a response indicating success is returned.

        Assertion is made by the response_handler.
        """
        api.Client(self.cfg, api.json_handler).get(
            ARTIFACTS_PATH, auth=HTTPBasicAuth(*self.cfg.pulp_auth)
        )

    def test_base_auth_failure(self):
        """Perform HTTP basic authentication with invalid credentials.

        Assert that a response indicating failure is returned.
        """
        self.cfg.pulp_auth[1] = utils.uuid4()  # randomize password
        response = api.Client(self.cfg, api.echo_handler).get(
            ARTIFACTS_PATH, auth=HTTPBasicAuth(*self.cfg.pulp_auth)
        )
        with self.assertRaises(HTTPError):
            response.raise_for_status()
        for key in ("invalid", "username", "password"):
            self.assertIn(key, response.json()["detail"].lower(), response.json())



# Import Serializers in platform that are potentially useful to plugin writers
from pulpcore.app.serializers import (  # noqa
    AlternateContentSourceSerializer,
    AlternateContentSourcePathSerializer,
    ArtifactSerializer,
    AsyncOperationResponseSerializer,
    ContentChecksumSerializer,
    ContentGuardSerializer,
    DetailRelatedField,
    DistributionSerializer,
    ExporterSerializer,
    ExportSerializer,
    IdentityField,
    ImporterSerializer,
    ImportSerializer,
    LabelsField,
    ModelSerializer,
    MultipleArtifactContentSerializer,
    NestedRelatedField,
    NoArtifactContentSerializer,
    ProgressReportSerializer,
    PublicationSerializer,
    RelatedField,
    RemoteSerializer,
    RepositorySerializer,
    RepositorySyncURLSerializer,
    RepositoryVersionRelatedField,
    SingleArtifactContentSerializer,
    SingleContentArtifactField,
    TaskGroupOperationResponseSerializer,
    RepositoryAddRemoveContentSerializer,
    ValidateFieldsMixin,
    validate_unknown_fields,
)

from .content import (  # noqa
    NoArtifactContentUploadSerializer,
    SingleArtifactContentUploadSerializer,
)



# Allow plugin viewsets to return 202s
from pulpcore.app.response import OperationPostponedResponse, TaskGroupOperationResponse  # noqa

# Import Viewsets in platform that are potentially useful to plugin writers
from pulpcore.app.viewsets import (  # noqa
    AlternateContentSourceViewSet,
    AsyncUpdateMixin,
    BaseFilterSet,
    ContentFilter,
    ContentGuardFilter,
    ContentGuardViewSet,
    ContentViewSet,
    DistributionFilter,
    DistributionViewSet,
    ExporterViewSet,
    ExportViewSet,
    ImmutableRepositoryViewSet,
    ImporterViewSet,
    ImportViewSet,
    NamedModelViewSet,
    NAME_FILTER_OPTIONS,
    PublicationFilter,
    PublicationViewSet,
    ReadOnlyContentViewSet,
    ReadOnlyRepositoryViewSet,
    RemoteFilter,
    RemoteViewSet,
    RepositoryVersionViewSet,
    RepositoryViewSet,
    RolesMixin,
    TaskGroupViewSet,
    TaskViewSet,
)

from pulpcore.app.viewsets.custom_filters import (  # noqa
    CharInFilter,
    HyperlinkRelatedFilter,
    LabelSelectFilter,
    RepositoryVersionFilter,
)

from .content import (  # noqa
    NoArtifactContentUploadViewSet,
    SingleArtifactContentUploadViewSet,
)



# Generated by Django 3.2.6 on 2021-08-25 16:55

from django.db import migrations, models
import django.db.models.deletion
import django_lifecycle.mixins
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0073_encrypt_remote_fields'),
    ]

    operations = [
        migrations.CreateModel(
            name='AlternateContentSource',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('pulp_type', models.TextField(db_index=True, default=None)),
                ('name', models.TextField(db_index=True, unique=True)),
                ('last_refreshed', models.DateTimeField(null=True)),
                ('remote', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='core.remote')),
            ],
            options={
                'verbose_name_plural': 'acs',
            },
            bases=(django_lifecycle.mixins.LifecycleModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='AlternateContentSourcePath',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('path', models.TextField(default=None)),
                ('alternate_content_source', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='paths', to='core.alternatecontentsource')),
                ('repository', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='core.repository')),
            ],
            options={
                'unique_together': {('alternate_content_source', 'path')},
            },
            bases=(django_lifecycle.mixins.LifecycleModelMixin, models.Model),
        ),
    ]



from gettext import gettext as _

from django.db import models
from rest_framework import mixins, permissions, status
from rest_framework.response import Response

from pulpcore.app.models import Artifact, Content, PublishedMetadata, SigningService
from pulpcore.app.serializers import (
    ArtifactSerializer,
    MultipleArtifactContentSerializer,
    SigningServiceSerializer,
)
from pulpcore.app.viewsets.base import BaseFilterSet, NamedModelViewSet

from .custom_filters import (
    ArtifactRepositoryVersionFilter,
    ContentAddedRepositoryVersionFilter,
    ContentRemovedRepositoryVersionFilter,
    ContentRepositoryVersionFilter,
)


class ArtifactFilter(BaseFilterSet):
    """
    Artifact filter Plugin content filters should:
     - inherit from this class
     - add any plugin-specific filters if needed
     - define its own `Meta` class should:
       - specify plugin content model
       - extend `fields` with plugin-specific ones
    """

    repository_version = ArtifactRepositoryVersionFilter()

    class Meta:
        model = Artifact
        fields = {
            "md5",
            "sha1",
            "sha224",
            "sha256",
            "sha384",
            "sha512",
        }


class ArtifactViewSet(
    NamedModelViewSet,
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    mixins.DestroyModelMixin,
):
    endpoint_name = "artifacts"
    queryset = Artifact.objects.all()
    serializer_class = ArtifactSerializer
    filterset_class = ArtifactFilter
    permission_classes = (permissions.IsAuthenticated,)

    def destroy(self, request, pk):
        """
        Remove Artifact only if it is not associated with any Content.
        """
        try:
            return super().destroy(request, pk)
        except models.ProtectedError:
            msg = _("The Artifact cannot be deleted because it is associated with Content.")
            data = {"detail": msg}
            return Response(data, status=status.HTTP_409_CONFLICT)


class ContentFilter(BaseFilterSet):
    """
    Plugin content filters should:
     - inherit from this class
     - add any plugin-specific filters if needed
     - define its own `Meta` class which should:
       - specify plugin content model
       - extend `fields` with plugin-specific ones

    Allows you to filter the content app by repository version.

    Fields:

        repository_version:
            Return Content which is contained within this repository version.
        repository_version_added:
            Return Content which was added in this repository version.
        repository_version_removed:
            Return Content which was removed from this repository version.
    """

    repository_version = ContentRepositoryVersionFilter()
    repository_version_added = ContentAddedRepositoryVersionFilter()
    repository_version_removed = ContentRemovedRepositoryVersionFilter()


class BaseContentViewSet(NamedModelViewSet):
    """
    A base class for any content viewset.

    It ensures that 'content/' is a part of endpoint. It also sets a default filter class.
    """

    endpoint_name = "content"
    filterset_class = ContentFilter
    # These are just placeholders, the plugin writer would replace them with the actual
    queryset = Content.objects.all().exclude(pulp_type=PublishedMetadata.get_pulp_type())
    serializer_class = MultipleArtifactContentSerializer


class ListContentViewSet(BaseContentViewSet, mixins.ListModelMixin):
    """Endpoint to list all content."""

    @classmethod
    def is_master_viewset(cls):
        """Do not hide from the routers."""
        return False


class ContentViewSet(
    BaseContentViewSet, mixins.CreateModelMixin, mixins.RetrieveModelMixin, mixins.ListModelMixin
):
    """
    Content viewset that supports POST and GET by default.
    """


class ReadOnlyContentViewSet(BaseContentViewSet, mixins.RetrieveModelMixin, mixins.ListModelMixin):
    """
    Content viewset that supports only GET by default.
    """


class SigningServiceViewSet(NamedModelViewSet, mixins.RetrieveModelMixin, mixins.ListModelMixin):
    """
    A ViewSet that supports browsing of existing signing services.
    """

    endpoint_name = "signing-services"
    queryset = SigningService.objects.all()
    serializer_class = SigningServiceSerializer
    filterset_fields = ["name"]



# Generated by Django 3.2.7 on 2021-09-17 10:23

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django_lifecycle.mixins
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('auth', '0012_alter_user_first_name_max_length'),
        ('core', '0077_move_remote_url_credentials'),
    ]

    operations = [
        migrations.CreateModel(
            name='Role',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('name', models.CharField(db_index=True, max_length=128, unique=True)),
                ('description', models.TextField(null=True)),
                ('permissions', models.ManyToManyField(to='auth.Permission')),
                ('locked', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
            },
            bases=(django_lifecycle.mixins.LifecycleModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='UserRole',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('object_id', models.CharField(max_length=255, null=True)),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
                ('role', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='object_users', to='core.role')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='object_roles', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'indexes': [models.Index(fields=['content_type', 'object_id'], name='core_userro_content_5c0477_idx')],
                'unique_together': {('user', 'role', 'content_type', 'object_id')},
            },
            bases=(django_lifecycle.mixins.LifecycleModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='GroupRole',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('object_id', models.CharField(max_length=255, null=True)),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='object_roles', to='auth.group')),
                ('role', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='object_groups', to='core.role')),
            ],
            options={
                'indexes': [models.Index(fields=['content_type', 'object_id'], name='core_groupr_content_ea7d37_idx')],
                'unique_together': {('group', 'role', 'content_type', 'object_id')},
            },
            bases=(django_lifecycle.mixins.LifecycleModelMixin, models.Model),
        ),
    ]



from gettext import gettext as _

from django_filters.rest_framework import filters
from drf_spectacular.utils import extend_schema
from rest_framework import mixins
from rest_framework.decorators import action

from pulpcore.app import tasks
from pulpcore.app.models import AlternateContentSource
from pulpcore.app.serializers import (
    AsyncOperationResponseSerializer,
    AlternateContentSourceSerializer,
)
from pulpcore.app.response import OperationPostponedResponse
from pulpcore.app.viewsets import (
    AsyncUpdateMixin,
    BaseFilterSet,
    NamedModelViewSet,
)
from pulpcore.app.viewsets.base import NAME_FILTER_OPTIONS
from pulpcore.tasking.tasks import dispatch


class AlternateContentSourceFilter(BaseFilterSet):
    """FilterSet for ACS."""

    name = filters.CharFilter()

    class Meta:
        model = AlternateContentSource
        fields = {
            "name": NAME_FILTER_OPTIONS,
        }


class AlternateContentSourceViewSet(
    mixins.CreateModelMixin,
    mixins.DestroyModelMixin,
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin,
    AsyncUpdateMixin,
    NamedModelViewSet,
):
    """
    A class for ACS viewset.
    """

    queryset = AlternateContentSource.objects.all()
    serializer_class = AlternateContentSourceSerializer
    endpoint_name = "acs"
    router_lookup = "acs"
    filterset_class = AlternateContentSourceFilter

    @action(detail=True, methods=["post"])
    def refresh(self, request, pk=None):
        raise NotImplementedError(_("Method not implemented by plugin writer!"))

    @extend_schema(
        description="Trigger an asynchronous delete ACS task",
        responses={202: AsyncOperationResponseSerializer},
    )
    def destroy(self, request, pk, **kwargs):
        acs = self.get_object()
        reservations = []
        instance_ids = []

        for path in acs.paths.all():
            if path.repository_id:
                instance_ids.append(
                    (str(path.repository_id), "core", "RepositorySerializer"),
                )
        reservations.append(acs)
        instance_ids.append(
            (str(acs.pk), "core", "AlternateContentSourceSerializer"),
        )
        async_result = dispatch(
            tasks.base.general_multi_delete, exclusive_resources=reservations, args=(instance_ids,)
        )
        return OperationPostponedResponse(async_result, request)



# Generated by Django 2.2.14 on 2020-07-15 14:44

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0035_content_upstream_id'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pulpexporter',
            name='last_export',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='last_export', to='core.PulpExport'),
        ),
    ]



"""Tests that look at generic list endpoints."""
import tempfile
import unittest

from pulp_smash import config, utils
from pulp_smash.pulp3.bindings import delete_orphans, monitor_task
from pulp_smash.pulp3.utils import gen_repo

from pulpcore.tests.functional.api.using_plugin.constants import X509_CA_CERT_FILE_PATH
from pulpcore.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

from pulpcore.client.pulpcore import (
    ApiClient as CoreApiClient,
    ContentApi,
    ContentguardsApi,
    RepositoriesApi,
)
from pulpcore.client.pulp_file import (
    ApiClient as FileApiClient,
    ContentFilesApi,
    RepositoriesFileApi,
)

from pulpcore.client.pulp_certguard import (
    ApiClient as CertGuardApiClient,
    ContentguardsX509Api,
)


class GenericListTestCase(unittest.TestCase):
    """Test generic list endpoints."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.file_repositories_api = RepositoriesFileApi(
            FileApiClient(cls.cfg.get_bindings_config())
        )
        cls.repo = cls.file_repositories_api.create(gen_repo())

        cls.file_content_api = ContentFilesApi(FileApiClient(cls.cfg.get_bindings_config()))
        with tempfile.NamedTemporaryFile() as tmp_file:
            tmp_file.write(b"not empty")
            tmp_file.flush()
            monitor_task(
                cls.file_content_api.create(relative_path=utils.uuid4(), file=tmp_file.name).task
            )

        cls.cert_guards_api = ContentguardsX509Api(
            CertGuardApiClient(cls.cfg.get_bindings_config())
        )
        with open(X509_CA_CERT_FILE_PATH, "r") as x509_ca_cert_data_file:
            x509_ca_cert_data = x509_ca_cert_data_file.read()

        cls.content_guard = cls.cert_guards_api.create(
            {"name": utils.uuid4(), "ca_certificate": x509_ca_cert_data}
        )

    @classmethod
    def tearDownClass(cls):
        """Cleanup class-wide variables."""
        monitor_task(cls.file_repositories_api.delete(cls.repo.pulp_href).task)
        cls.cert_guards_api.delete(cls.content_guard.pulp_href)
        delete_orphans()

    def test_read_all_repos_generic(self):
        """Ensure name is displayed when listing repositories generic."""
        repositories_api = RepositoriesApi(CoreApiClient(self.cfg.get_bindings_config()))

        response = repositories_api.list()
        self.assertNotEqual(response.count, 0)
        for repo in response.results:
            self.assertIsNotNone(repo.name)

    def test_read_all_content_generic(self):
        """Ensure href is displayed when listing content generic."""
        content_api = ContentApi(CoreApiClient(self.cfg.get_bindings_config()))

        response = content_api.list()
        self.assertNotEqual(response.count, 0)
        for content in response.results:
            self.assertIsNotNone(content.pulp_href)

    def test_read_all_content_guards_generic(self):
        """Ensure name is displayed when listing content guards generic."""
        content_guards_api = ContentguardsApi(CoreApiClient(self.cfg.get_bindings_config()))

        response = content_guards_api.list()
        self.assertNotEqual(response.count, 0)
        for content_guard in response.results:
            self.assertIsNotNone(content_guard.name)



# Generated by Django 2.2.17 on 2021-01-29 19:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0056_remote_rate_limit'),
    ]

    operations = [
        migrations.AlterField(
            model_name='label',
            name='key',
            field=models.CharField(db_index=True, max_length=200),
        ),
        migrations.AlterField(
            model_name='label',
            name='value',
            field=models.TextField(db_index=True, null=True),
        ),
    ]



# Generated by Django 2.2.15 on 2020-08-25 19:32

from django.db import migrations
import pulpcore.app.models.content
import pulpcore.app.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0043_toc_attribute'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pulptemporaryfile',
            name='file',
            field=pulpcore.app.models.fields.ArtifactFileField(max_length=255, upload_to=pulpcore.app.models.content.PulpTemporaryFile.storage_path),
        ),
    ]



from django.http import Http404
from django_filters.rest_framework import filters
from drf_spectacular.utils import extend_schema
from rest_framework import mixins

from pulpcore.app.models import (
    Import,
    Importer,
    PulpImport,
    PulpImporter,
    TaskGroup,
)
from pulpcore.app.response import TaskGroupOperationResponse
from pulpcore.app.serializers import (
    ImportSerializer,
    ImporterSerializer,
    PulpImporterSerializer,
    PulpImportSerializer,
    TaskGroupOperationResponseSerializer,
)
from pulpcore.app.tasks import pulp_import
from pulpcore.app.viewsets import (
    BaseFilterSet,
    NamedModelViewSet,
)
from pulpcore.app.viewsets.base import NAME_FILTER_OPTIONS
from pulpcore.tasking.tasks import dispatch


class ImporterFilter(BaseFilterSet):
    """Filter for Importers."""

    name = filters.CharFilter()

    class Meta:
        model = Importer
        fields = {
            "name": NAME_FILTER_OPTIONS,
        }


class ImporterViewSet(
    NamedModelViewSet,
    mixins.CreateModelMixin,
    mixins.UpdateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    mixins.DestroyModelMixin,
):
    """ViewSet for Importers."""

    queryset = Importer.objects.all()
    serializer_class = ImporterSerializer
    endpoint_name = "importers"
    router_lookup = "importer"
    filterset_class = ImporterFilter


class ImportViewSet(
    NamedModelViewSet,
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    mixins.DestroyModelMixin,
):
    """ViewSet for viewing imports from an Importer."""

    endpoint_name = "imports"
    nest_prefix = "importers"
    router_lookup = "import"
    lookup_field = "pk"
    parent_viewset = ImporterViewSet
    parent_lookup_kwargs = {"importer_pk": "importer__pk"}
    serializer_class = ImportSerializer
    queryset = Import.objects.all()


class PulpImporterViewSet(ImporterViewSet):
    """ViewSet for PulpImporters."""

    endpoint_name = "pulp"
    serializer_class = PulpImporterSerializer
    queryset = PulpImporter.objects.all()


class PulpImportViewSet(ImportViewSet):
    """ViewSet for PulpImports."""

    parent_viewset = PulpImporterViewSet
    queryset = PulpImport.objects.all()

    @extend_schema(
        request=PulpImportSerializer,
        description="Trigger an asynchronous task to import a Pulp export.",
        responses={202: TaskGroupOperationResponseSerializer},
    )
    def create(self, request, importer_pk):
        """Import a Pulp export into Pulp."""
        try:
            importer = PulpImporter.objects.get(pk=importer_pk)
        except PulpImporter.DoesNotExist:
            raise Http404

        serializer = PulpImportSerializer(data=request.data, context={"request": request})
        serializer.is_valid(raise_exception=True)

        path = serializer.validated_data.get("path")
        toc = serializer.validated_data.get("toc")
        task_group = TaskGroup.objects.create(description=f"Import of {path}")

        dispatch(
            pulp_import,
            exclusive_resources=[importer],
            task_group=task_group,
            kwargs={"importer_pk": importer.pk, "path": path, "toc": toc},
        )
        return TaskGroupOperationResponse(task_group, request)



from gettext import gettext as _

from django.core.exceptions import ImproperlyConfigured
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group as BaseGroup
from django.db import models
from django_currentuser.middleware import get_current_authenticated_user
from django_lifecycle import hook, LifecycleModelMixin
from guardian.models.models import GroupObjectPermission, UserObjectPermission
from guardian.shortcuts import assign_perm, remove_perm

from pulpcore.app.models import BaseModel
from pulpcore.app.loggers import deprecation_logger


def _ensure_iterable(obj):
    if isinstance(obj, str):
        return [obj]
    return obj


class AccessPolicy(BaseModel):
    """
    A model storing a viewset authorization policy and permission assignment of new objects created.

    Fields:

        creation_hooks (models.JSONField): A list of dictionaries identifying callables on the
            ``pulpcore.plugin.access_policy.AccessPolicyFromDB`` which can add user or group roles
            for newly created objects. This is a nullable field due to not all endpoints creating
            objects.
        statements (models.JSONField): A list of ``drf-access-policy`` statements.
        viewset_name (models.CharField): The name of the viewset this instance controls
            authorization for.
        customized (BooleanField): False if the AccessPolicy has been user-modified. True otherwise.
            Defaults to False.

    """

    creation_hooks = models.JSONField(null=True)
    statements = models.JSONField()
    viewset_name = models.CharField(max_length=128, unique=True)
    customized = models.BooleanField(default=False)


class AutoAddObjPermsMixin:
    """
    A mixin that automatically adds roles based on the ``creation_hooks`` data.

    To use this mixin, your model must support ``django-lifecycle``.

    To use this mixin, you must define a class attribute named ``ACCESS_POLICY_VIEWSET_NAME``
    containing the name of the ViewSet associated with this object.

    This mixin adds an ``after_create`` hook which properly interprets the ``creation_hooks``
    data and calls methods also provided by this mixin to add roles.

    These mixing are provided by default:

    * ``add_for_object_creator`` will add the permissions to the creator of the object.
    * ``add_for_users`` will add the permissions for one or more users by name.
    * ``add_for_groups`` will add the permissions for one or more groups by name.
    * ``add_roles_for_object_creator`` will add the roles to the creator of the object.
    * ``add_roles_for_users`` will add the roles for one or more users by name.
    * ``add_roles_for_groups`` will add the roles for one or more groups by name.

    """

    def __init__(self, *args, **kwargs):
        self.REGISTERED_CREATION_HOOKS = {
            "add_roles_for_users": self.add_roles_for_users,
            "add_roles_for_groups": self.add_roles_for_groups,
            "add_roles_for_object_creator": self.add_roles_for_object_creator,
        }
        super().__init__(*args, **kwargs)

    @hook("after_create")
    def add_perms(self):
        try:
            access_policy = AccessPolicy.objects.get(viewset_name=self.ACCESS_POLICY_VIEWSET_NAME)
        except AttributeError:
            raise ImproperlyConfigured(
                _(
                    "When using the `AutoAddObjPermsMixin`, plugin writers must declare an"
                    "`ACCESS_POLICY_VIEWSET_NAME` class attribute."
                )
            )
        self._handle_creation_hooks(access_policy)

    def _handle_creation_hooks(self, access_policy):
        if access_policy.creation_hooks is not None:
            for creation_hook in access_policy.creation_hooks:
                function = self.REGISTERED_CREATION_HOOKS.get(creation_hook["function"])
                if function is not None:
                    kwargs = creation_hook.get("parameters") or {}
                    function(**kwargs)
                else:
                    # Old interface deprecated for removal in 3.20
                    function = getattr(self, creation_hook["function"])
                    deprecation_logger.warn(
                        "Calling unregistered creation hooks from the access policy is deprecated"
                        " and may be removed with pulpcore 3.20 "
                        f"[hook={creation_hook}, viewset={access_policy.viewset_name}]."
                    )
                    function(creation_hook.get("permissions"), creation_hook.get("parameters"))

    def add_for_users(self, permissions, users):
        """
        Adds object-level permissions for one or more users for this newly created object.

        Args:
            permissions (str or list): One or more object-level permissions to be added for the
                users. The permission names are in the form of ``<app_name>.<codename>``, e.g.
                ``"core.change_task"``. This can either be a single permission as a string, or list
                of permission names.
            users (str or list): One or more users who will receive object-level permissions. This
                can either be a single username as a string or a list of usernames.

        Raises:
            ObjectDoesNotExist: If any of the users do not exist.

        """
        permissions = _ensure_iterable(permissions)
        users = _ensure_iterable(users)
        for username in users:
            user = get_user_model().objects.get(username=username)
            for perm in permissions:
                assign_perm(perm, user, self)

    def add_roles_for_users(self, roles, users):
        """
        Adds object-level roles for one or more users for this newly created object.

        Args:
            roles (str or list): One or more roles to be added at object-level for the users.
                This can either be a single role as a string, or a list of role names.
            users (str or list): One or more users who will receive object-level roles. This can
                either be a single username as a string or a list of usernames.

        Raises:
            ObjectDoesNotExist: If any of the users do not exist.

        """
        from pulpcore.app.role_util import assign_role

        roles = _ensure_iterable(roles)
        users = _ensure_iterable(users)
        for username in users:
            user = get_user_model().objects.get(username=username)
            for role in roles:
                assign_role(role, user, self)

    def add_for_groups(self, permissions, groups):
        """
        Adds object-level permissions for one or more groups for this newly created object.

        Args:
            permissions (str or list): One or more object-level permissions to be added for the
                groups. The permission names are in the form of ``<app_name>.<codename>``, e.g.
                ``"core.change_task"``. This can either be a single permission as a string, or list
                of permission names.
            groups (str or list): One or more groups who will receive object-level permissions. This
                can either be a single group name as a string or a list of group names.

        Raises:
            ObjectDoesNotExist: If any of the groups do not exist.

        """
        permissions = _ensure_iterable(permissions)
        groups = _ensure_iterable(groups)
        for group_name in groups:
            group = Group.objects.get(name=group_name)
            for perm in permissions:
                assign_perm(perm, group, self)

    def add_roles_for_groups(self, roles, groups):
        """
        Adds object-level roles for one or more groups for this newly created object.

        Args:
            roles (str or list): One or more object-level roles to be added for the groups. This
                can either be a single role as a string, or list of role names.
            groups (str or list): One or more groups who will receive object-level roles. This
                can either be a single group name as a string or a list of group names.

        Raises:
            ObjectDoesNotExist: If any of the groups do not exist.

        """
        from pulpcore.app.role_util import assign_role

        roles = _ensure_iterable(roles)
        groups = _ensure_iterable(groups)
        for group_name in groups:
            group = Group.objects.get(name=group_name)
            for role in roles:
                assign_role(role, group, self)

    def add_for_object_creator(self, permissions, *args):
        """
        Adds object-level permissions for the user creating the newly created object.

        If the ``django_currentuser.middleware.get_current_authenticated_user`` returns None because
        the API client did not provide authentication credentials, *no* permissions are added and
        this passes silently. This allows endpoints which create objects and do not require
        authorization to execute without error.

        Args:
            permissions (str or list): One or more object-level permissions to be added for the
                users. The permission names are in the form of ``<app_name>.<codename>``, e.g.
                ``"core.change_task"``. This can either be a single permission as a string, or list
                of permission names.

        """
        permissions = _ensure_iterable(permissions)
        current_user = get_current_authenticated_user()
        if current_user:
            for perm in permissions:
                assign_perm(perm, current_user, self)

    def add_roles_for_object_creator(self, roles):
        """
        Adds object-level roles for the user creating the newly created object.

        If the ``django_currentuser.middleware.get_current_authenticated_user`` returns None because
        the API client did not provide authentication credentials, *no* permissions are added and
        this passes silently. This allows endpoints which create objects and do not require
        authorization to execute without error.

        Args:
            roles (list or str): One or more roles to be added at the object-level for the user.
                This can either be a single role as a string, or list of role names.

        """
        from pulpcore.app.role_util import assign_role

        roles = _ensure_iterable(roles)
        current_user = get_current_authenticated_user()
        if current_user:
            for role in roles:
                assign_role(role, current_user, self)


class AutoDeleteObjPermsMixin:
    """
    A mixin that automatically deletes user and group permissions for an object prior to deletion.

    To use this mixin, your model must support ``django-lifecycle``.
    """

    @hook("before_delete")
    def delete_user_and_group_obj_perms(self):
        UserObjectPermission.objects.filter(object_pk=self.pk).delete()
        GroupObjectPermission.objects.filter(object_pk=self.pk).delete()

    def remove_for_users(self, permissions, users):
        """
        Removes object-level permissions for one or more users for this object.
        Requires AutoAddObjPermsMixin

        Args:
            permissions (str or list): One or more object-level permissions to be added for the
                users. The permission names are in the form of ``<app_name>.<codename>``, e.g.
                ``"core.change_task"``. This can either be a single permission as a string, or list
                of permission names.
            users (str or list): One or more users who will lose these object-level permissions.
                This can either be a single username as a string or a list of usernames.

        Raises:
            ObjectDoesNotExist: If any of the users do not exist.

        """
        permissions = _ensure_iterable(permissions)
        users = _ensure_iterable(users)
        for username in users:
            user = get_user_model().objects.get(username=username)
            for perm in permissions:
                remove_perm(perm, user, self)

    def remove_for_groups(self, permissions, groups):
        """
        Removes object-level permissions for one or more groups for this object.
        Requires AutoAddObjPermsMixin

        Args:
            permissions (str or list): One or more object-level permissions to be removed for the
                groups. The permission names are in the form of ``<app_name>.<codename>``, e.g.
                ``"core.change_task"``. This can either be a single permission as a string, or list
                of permission names.
            groups (str or list): One or more groups who will lose these object-level permissions.
                This can either be a single group name as a string or a list of group names.

        Raises:
            ObjectDoesNotExist: If any of the groups do not exist.

        """
        permissions = _ensure_iterable(permissions)
        groups = _ensure_iterable(groups)
        for group_name in groups:
            group = Group.objects.get(name=group_name)
            for perm in permissions:
                remove_perm(perm, group, self)


class Group(LifecycleModelMixin, BaseGroup, AutoAddObjPermsMixin):
    ACCESS_POLICY_VIEWSET_NAME = "groups"

    class Meta:
        proxy = True
        permissions = [
            ("manage_roles_group", "Can manage role assignments on group"),
        ]



import os
import re
from gettext import gettext as _

from django.conf import settings
from rest_framework import serializers
from rest_framework.fields import empty
from rest_framework.reverse import reverse

from pulpcore.app import models
from pulpcore.app.serializers import DetailIdentityField, IdentityField, RelatedField


def relative_path_validator(relative_path):
    if os.path.isabs(relative_path):
        raise serializers.ValidationError(
            _("Relative path can't start with '/'. {0}").format(relative_path)
        )


class SingleContentArtifactField(RelatedField):
    """
    A serializer field for the '_artifacts' ManyToManyField on the Content model (single-artifact).
    """

    lookup_field = "pk"
    view_name = "artifacts-detail"
    queryset = models.Artifact.objects.all()
    allow_null = True

    def get_attribute(self, instance):
        """
        Returns the field from the instance that should be serialized using this serializer field.

        This serializer looks up the list of artifacts and returns only one, if any exist. If more
        than one exist, it throws and exception because this serializer is being used in an
        improper context.

        Args:
            instance (:class:`pulpcore.app.models.Content`): An instance of Content being
                serialized.

        Returns:
            A single Artifact model related to the instance of Content.
        """
        # using get() and first() will query the db. count() and all() will use cached artifacts if
        # they are prefetched.
        if instance._artifacts.count() == 0:
            return None
        if instance._artifacts.count() == 1:
            return instance._artifacts.all()[0]
        if instance._artifacts.count() > 1:
            raise ValueError(
                _(
                    "SingleContentArtifactField should not be used in a context where "
                    "multiple artifacts for one content is possible."
                )
            )


class ContentArtifactChecksumField(serializers.CharField):
    """
    A serializer field for the artifact checksum Content model (single-artifact).
    """

    def __init__(self, *args, **kwargs):
        kwargs["read_only"] = True
        self.checksum = kwargs.pop("checksum", "sha256")
        super().__init__(*args, **kwargs)

    def get_attribute(self, instance):
        """
        Returns the field from the instance that should be serialized using this serializer field.

        This serializer looks up the checksum for single artifact content

        Args:
            instance (:class:`pulpcore.app.models.Content`): An instance of Content being
                serialized.

        Returns:
            A string of the checksum or None.

        Raises:
            :class:`rest_framework.exceptions.ValidationError`: When more than one Artifacts exist.
        """
        # using get() and first() will query the db. count() and all() will use cached artifacts if
        # they are prefetched.
        if instance._artifacts.count() == 0:
            return None
        if instance._artifacts.count() == 1:
            return getattr(instance._artifacts.all()[0], self.checksum)
        if instance._artifacts.count() > 1:
            raise ValueError(
                _(
                    "ContentArtifactChecksumField should not be used in a context where "
                    "multiple artifacts for one content is possible."
                )
            )


class ContentArtifactsField(serializers.DictField):
    """
    A serializer field for the '_artifacts' ManyToManyField on the Content model.
    """

    def run_validation(self, data):
        """
        Validates 'data' dict.

        Validates that all keys of 'data' are relative paths. Validates that all values of 'data'
        are URLs for an existing Artifact.

        Args:
            data (dict): A dict mapping relative paths inside the Content to the corresponding
                Artifact URLs.

        Returns:
            A dict mapping relative paths inside the Content to the corresponding Artifact
                instances.

        Raises:
            :class:`rest_framework.exceptions.ValidationError`: When one of the Artifacts does not
                exist or one of the paths is not a relative path or the field is missing.
        """
        ret = {}
        if data is empty:
            raise serializers.ValidationError(_("artifacts field must be specified."))
        for relative_path, url in data.items():
            relative_path_validator(relative_path)
            artifactfield = RelatedField(
                view_name="artifacts-detail",
                queryset=models.Artifact.objects.all(),
                source="*",
                initial=url,
            )
            try:
                artifact = artifactfield.run_validation(data=url)
                ret[relative_path] = artifact
            except serializers.ValidationError as e:
                # Append the URL of missing Artifact to the error message
                e.detail[0] = "%s %s" % (e.detail[0], url)
                raise e
        return ret

    def get_attribute(self, instance):
        """
        Returns the field from the instance that should be serialized using this serializer field.

        This serializer field serializes a ManyToManyField that is actually stored as a
        ContentArtifact model. Instead of returning the field, this method returns all the
        ContentArtifact models related to this Content.

        Args:
            instance (:class:`pulpcore.app.models.Content`): An instance of Content being
                serialized.

        Returns:
            A list of ContentArtifact models related to the instance of Content.
        """
        return instance.contentartifact_set.all()

    def to_representation(self, value):
        """
        Serializes list of ContentArtifacts.

        Returns a dict mapping relative paths inside the Content to the corresponding Artifact
        URLs.

        Args:
            value (list of :class:`pulpcore.app.models.ContentArtifact`): A list of all the
                ContentArtifacts related to the Content model being serialized.

        Returns:
            A dict where keys are relative path of the artifact inside the Content and values are
                Artifact URLs.
        """
        ret = {}
        for content_artifact in value:
            if content_artifact.artifact_id:
                url = reverse(
                    "artifacts-detail", kwargs={"pk": content_artifact.artifact_id}, request=None
                )
            else:
                url = None
            ret[content_artifact.relative_path] = url
        return ret


class RepositoryVersionsIdentityFromRepositoryField(DetailIdentityField):
    view_name = "repositories-detail"

    def __init__(self, view_name=None, **kwargs):
        assert view_name is None, "The `view_name` must not be set."
        super().__init__(view_name=self.view_name, **kwargs)

    def get_url(self, obj, view_name, request, *args, **kwargs):
        return super().get_url(obj, self.view_name, request, *args, **kwargs) + "versions/"


class RepositoryVersionFieldGetURLMixin:
    view_name = "versions-detail"

    def __init__(self, view_name=None, **kwargs):
        assert view_name is None, "The `view_name` must not be set."
        super().__init__(view_name=self.view_name, **kwargs)

    def get_url(self, obj, view_name, request, *args, **kwargs):
        rvr_field = RepositoryVersionsIdentityFromRepositoryField()
        repo_url = rvr_field.get_url(obj.repository, None, request, *args, **kwargs)
        return f"{repo_url}{obj.number}/"

    def use_pk_only_optimization(self):
        return False


class RepositoryVersionIdentityField(RepositoryVersionFieldGetURLMixin, IdentityField):
    pass


class RepositoryVersionRelatedField(RepositoryVersionFieldGetURLMixin, RelatedField):
    queryset = models.RepositoryVersion.objects.all()

    def get_object(self, view_name, view_args, view_kwargs):
        lookup_kwargs = {
            "repository__pk": view_kwargs["repository_pk"],
            "number": view_kwargs["number"],
        }
        return self.get_queryset().get(**lookup_kwargs)


class LatestVersionField(RepositoryVersionRelatedField):
    queryset = None  # read-only relational fields should not provide a `queryset` argument

    def __init__(self, *args, **kwargs):
        """
        Unfortunately you can't just set read_only=True on the class. It has
        to be done explicitly in the kwargs to __init__, or else DRF complains.
        """
        kwargs["read_only"] = True
        super().__init__(*args, **kwargs)

    def get_url(self, obj, view_name, request, format):
        """
        Returns the URL for the appropriate object. Overrides the DRF method to change how the
        object is found.

        Args:
            obj (list of :class:`pulpcore.app.models.RepositoryVersion`): The versions of the
                current viewset's repository.
            view_name (str): The name of the view that should be used.
            request (rest_framework.request.Request): the current HTTP request being handled
            format: undocumented by DRF. ???

        Returns:
            str: the URL corresponding to the latest version of the current repository. If there
                are no versions, returns None
        """
        try:
            version = obj.complete().latest()
        except obj.model.DoesNotExist:
            return None
        return super().get_url(version, view_name, request, format)

    def get_attribute(self, instance):
        """
        Args:
            instance (pulpcore.app.models.Repository): a repository that has been matched by the
                current ViewSet.

        Returns:
            list of :class:`pulpcore.app.models.RepositoryVersion`

        """
        return instance.versions


class BaseURLField(serializers.CharField):
    """
    Serializer Field for the base_url field of the Distribution.
    """

    def to_representation(self, value):
        origin = settings.CONTENT_ORIGIN.strip("/")
        prefix = settings.CONTENT_PATH_PREFIX.strip("/")
        base_path = value.strip("/")

        return "/".join((origin, prefix, base_path)) + "/"


class ExportsIdentityFromExporterField(DetailIdentityField):
    view_name = "exporters-detail"

    def __init__(self, view_name=None, **kwargs):
        assert view_name is None, "The `view_name` must not be set."
        super().__init__(view_name=self.view_name, **kwargs)

    def get_url(self, obj, view_name, request, *args, **kwargs):
        return super().get_url(obj, self.view_name, request, *args, **kwargs) + "exports/"


class ExportFieldGetURLMixin:
    view_name = "exports-detail"

    def __init__(self, view_name=None, **kwargs):
        assert view_name is None, "The `view_name` must not be set."
        super().__init__(view_name=self.view_name, **kwargs)

    def get_url(self, obj, view_name, request, *args, **kwargs):
        exports_field = ExportsIdentityFromExporterField()
        exporter_url = exports_field.get_url(obj.exporter, None, request, *args, **kwargs)
        return f"{exporter_url}{obj.pk}/"

    def use_pk_only_optimization(self):
        return False


class ExportIdentityField(ExportFieldGetURLMixin, IdentityField):
    pass


class ExportRelatedField(ExportFieldGetURLMixin, RelatedField):
    queryset = models.Export.objects.all()

    def get_object(self, view_name, view_args, view_kwargs):
        lookup_kwargs = {"exporter__pk": view_kwargs["exporter_pk"], "pk": view_kwargs["pk"]}
        return self.get_queryset().get(**lookup_kwargs)


class ImportsIdentityFromImporterField(DetailIdentityField):
    view_name = "importers-detail"

    def __init__(self, view_name=None, **kwargs):
        assert view_name is None, "The `view_name` must not be set."
        super().__init__(view_name=self.view_name, **kwargs)

    def get_url(self, obj, view_name, request, *args, **kwargs):
        return super().get_url(obj, self.view_name, request, *args, **kwargs) + "imports/"


class ImportFieldGetURLMixin:
    view_name = "imports-detail"

    def __init__(self, view_name=None, **kwargs):
        assert view_name is None, "The `view_name` must not be set."
        super().__init__(view_name=self.view_name, **kwargs)

    def get_url(self, obj, view_name, request, *args, **kwargs):
        imports_field = ImportsIdentityFromImporterField()
        importer_url = imports_field.get_url(obj.importer, None, request, *args, **kwargs)
        return f"{importer_url}{obj.pk}/"

    def use_pk_only_optimization(self):
        return False


class ImportIdentityField(ImportFieldGetURLMixin, IdentityField):
    pass


class ImportRelatedField(ImportFieldGetURLMixin, RelatedField):
    queryset = models.Import.objects.all()

    def get_object(self, view_name, view_args, view_kwargs):
        lookup_kwargs = {"importer__pk": view_kwargs["importer_pk"], "pk": view_kwargs["pk"]}
        return self.get_queryset().get(**lookup_kwargs)


class TaskGroupStatusCountField(serializers.IntegerField, serializers.ReadOnlyField):
    """Serializer field for counting the tasks on a task group in a given state."""

    def __init__(self, state, *args, **kwargs):
        self.state = state
        super().__init__(*args, **kwargs)

    def get_attribute(self, instance):
        return instance.tasks.filter(state=self.state).count()


class LabelsField(serializers.JSONField):
    """A serializer field for pulp_labels."""

    def to_representation(self, labels):
        """
        Serializes list of labels to a dict.

        Args:
            labels (list of :class:`pulpcore.app.models.Label`): A list of labels to serialize

        Returns:
            A dict that maps label keys to label values
        """
        return {label.key: label.value for label in labels.all()}

    def to_internal_value(self, data):
        """
        Ensures that data conforms to key/value dict.

        Args:
            data (str or dict): A dictionary which maps a label key to a label value

        Returns:
            A validated data dict for labels mapping keys to values

        Raises:
            rest_framework.serializers.ValidationError: if data is invalid (e.g., not a dict)
        """
        data = super().to_internal_value(data)

        if not isinstance(data, dict):
            raise serializers.ValidationError(_("Data must be supplied as a key/value hash."))
        for key, value in data.items():
            if not re.match(r"^[\w ]+$", key):
                raise serializers.ValidationError(
                    _("Key '{}' contains non-alphanumerics.").format(key)
                )
            if re.search(r"[,()]", value):
                raise serializers.ValidationError(
                    _("Key '{}' contains value with comma or parenthesis.").format(key)
                )

        return data



# Generated by Django 2.2.20 on 2021-05-07 18:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0067_add_protect_to_task_reservation'),
    ]

    operations = [
        migrations.AddField(
            model_name='artifact',
            name='timestamp_of_interest',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name='content',
            name='timestamp_of_interest',
            field=models.DateTimeField(auto_now=True),
        ),
    ]



import http.client
from gettext import gettext as _

from .base import PulpException


class MissingResource(PulpException):
    """
    Base class for missing resource exceptions.

    Exceptions that are raised due to requests for resources that do not exist should inherit
    from this base class.
    """

    http_status_code = http.client.NOT_FOUND

    def __init__(self, **resources):
        """
        :param resources: keyword arguments of resource_type=resource_id
        :type resources: dict
        """
        super().__init__("PLP0001")
        self.resources = resources

    def __str__(self):
        resources_str = ", ".join("%s=%s" % (k, v) for k, v in self.resources.items())
        msg = _("The following resources are missing: %s") % resources_str
        return msg.encode("utf-8")



import unittest

from pulp_smash import config
from pulp_smash.pulp3.bindings import delete_orphans, monitor_task

from pulp_smash.pulp3.utils import (
    gen_repo,
    get_content,
)

from pulpcore.client.pulp_file import (
    FileFilePublication,
    PublicationsFileApi,
    RemotesFileApi,
    RepositoriesFileApi,
    RepositorySyncURL,
)
from pulpcore.client.pulp_file.exceptions import ApiException

from pulpcore.tests.functional.api.using_plugin.constants import (
    FILE_MANY_FIXTURE_MANIFEST_URL,
    FILE_CONTENT_NAME,
)
from pulpcore.tests.functional.api.using_plugin.utils import (
    gen_file_client,
    gen_file_remote,
)


class ContentInPublicationViewTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.file_client = gen_file_client()
        cls.repo_api = RepositoriesFileApi(cls.file_client)
        cls.publication_api = PublicationsFileApi(cls.file_client)
        cls.remote_api = RemotesFileApi(cls.file_client)

    @classmethod
    def tearDownClass(cls):
        delete_orphans()

    def test_all(self):
        """Create two publications and check view filter."""
        repo = self.repo_api.create(gen_repo())
        self.addCleanup(self.repo_api.delete, repo.pulp_href)

        remote = self.remote_api.create(gen_file_remote())
        self.addCleanup(self.remote_api.delete, remote.pulp_href)

        # Sync and update repository data.
        repo_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = self.repo_api.sync(repo.pulp_href, repo_sync_data)
        monitor_task(sync_response.task)
        repo = self.repo_api.read(repo.pulp_href)

        # Test content doesn't exists.
        non_existant_content_href = (
            "/pulp/api/v3/content/file/files/c4ed74cf-a806-490d-a25f-94c3c3dd2dd7/"
        )
        with self.assertRaises(ApiException) as ctx:
            self.publication_api.list(content=non_existant_content_href)
        self.assertEqual(ctx.exception.status, 400)

        # Test not published content.
        content_href = get_content(repo.to_dict())[FILE_CONTENT_NAME][0]["pulp_href"]
        self.assertEqual(self.publication_api.list(content=content_href).to_dict()["count"], 0)

        # Publication
        publication_data = FileFilePublication(repository=repo.pulp_href)
        publication_response = self.publication_api.create(publication_data)
        task_response = monitor_task(publication_response.task)
        publication = self.publication_api.read(task_response.created_resources[0])
        self.addCleanup(self.publication_api.delete, publication.pulp_href)

        # Second publication
        repo_second = self.repo_api.create(gen_repo())
        self.addCleanup(self.repo_api.delete, repo_second.pulp_href)

        body = gen_file_remote(url=FILE_MANY_FIXTURE_MANIFEST_URL)
        remote_second = self.remote_api.create(body)
        self.addCleanup(self.remote_api.delete, remote_second.pulp_href)

        repo_second_sync_data = RepositorySyncURL(remote=remote_second.pulp_href)
        sync_response = self.repo_api.sync(repo_second.pulp_href, repo_second_sync_data)
        monitor_task(sync_response.task)
        repo_second = self.repo_api.read(repo_second.pulp_href)

        publication_data = FileFilePublication(repository=repo_second.pulp_href)
        publication_response = self.publication_api.create(publication_data)
        task_response = monitor_task(publication_response.task)
        publication_second = self.publication_api.read(task_response.created_resources[0])
        self.addCleanup(self.publication_api.delete, publication_second.pulp_href)

        # Test there are two publications
        self.assertEqual(self.publication_api.list().count, 2)

        # Test content match publication
        self.assertEqual(self.publication_api.list(content=content_href).count, 1)
        self.assertEqual(
            self.publication_api.list(content=content_href).results[0].repository_version,
            repo.latest_version_href,
        )
